import json
import logging
import time

import pytest

from TradeGates.TradeGate import TradeGate

loglevel = logging.INFO
logging.basicConfig(level=loglevel)
log = logging.getLogger(__name__)


@pytest.fixture
def getGatesAndSymbolNames():
    gates = []
    symbolNames = {}
    with open("../../config.json") as f:
        config = json.load(f)

    for key in config.keys():
        gates.append(TradeGate(config[key], sandbox=True))
        if gates[-1].exchangeName.lower() == "kucoin":
            symbolNames[gates[-1].exchangeName] = "BTC-USDT"
        else:
            symbolNames[gates[-1].exchangeName] = "BTCUSDT"

    return gates, symbolNames


def testNewTestOrder(getGatesAndSymbolNames):
    gates, symbolNamesDict = getGatesAndSymbolNames
    for gate in gates:
        symbolName = symbolNamesDict[gate.exchangeName]
        try:
            res = gate.create_and_test_spot_order(
                symbolName,
                "SELL",
                "LIMIT",
                time_in_force="GTC",
                quantity=0.002,
                price=20000,
            )
            assert (
                res is not None
            ), "Problem in testing making order from {} exchange".format(
                gate.exchangeName
            )
        except Exception as e:
            assert False, "From {} exchange : {}".format(gate.exchangeName, str(e))


def testNewTestOrderBadOrderType(getGatesAndSymbolNames):
    gates, symbolNamesDict = getGatesAndSymbolNames
    for gate in gates:
        symbolName = symbolNamesDict[gate.exchangeName]
        try:
            res = gate.create_and_test_spot_order(
                symbolName,
                "SELL",
                "LINIT",
                time_in_force="GTC",
                quantity=0.002,
                price=30000,
            )
            assert (
                res is None
            ), "Bad order type and information provided. Must fail (Exchange: {})".format(
                gate.exchangeName
            )
        except Exception as e:
            assert (
                True
            ), "Bad order type and information provided. Must fail (Exchange: {})".format(
                gate.exchangeName
            )


def testNewOrder(getGatesAndSymbolNames):
    gates, symbolNamesDict = getGatesAndSymbolNames
    for gate in gates:
        symbolName = symbolNamesDict[gate.exchangeName]
        try:
            verifiedOrder = gate.create_and_test_spot_order(
                symbolName,
                "BUY",
                "LIMIT",
                quantity=0.002,
                price=20000,
                time_in_force="GTC",
            )
            result = gate.make_spot_order(verifiedOrder)
            print(result)
            assert (
                result is not None
            ), "Problem in making new order in {} exchange".format(gate.exchangeName)
        except Exception:
            assert False, "Problem in making new order in {} exchange".format(
                gate.exchangeName
            )


def testGetOrders(getGatesAndSymbolNames):
    gates, symbolNamesDict = getGatesAndSymbolNames
    for gate in gates:
        if gate.exchangeName.lower() == "binance":
            continue
        symbolName = symbolNamesDict[gate.exchangeName]
        orders = gate.get_symbol_orders(symbolName, futures=False)
        # print('\nGetting order history for BTCUSDT symbol from {}: {}'.format(gate.exchangeName, orders[0]))

        assert (
            orders is not None
        ), "Problem in getting list of all orders from {} exchange.".format(
            gate.exchangeName
        )


def testGetOpenOrders(getGatesAndSymbolNames):
    gates, symbolNamesDict = getGatesAndSymbolNames
    for gate in gates:
        if gate.exchangeName.lower() == "binance":
            continue
        symbolName = symbolNamesDict[gate.exchangeName]
        symbolOpenOrders = gate.get_open_orders(symbolName)
        # print('\nGetting BTCUSDT open orders list from {} exchange: {}'.format(gate.exchangeName, symbolOpenOrders))
        assert (
            symbolOpenOrders is not None
        ), "Problem in getting list of open orders with symbol from {} exchange.".format(
            gate.exchangeName
        )


def testGetOrder(getGatesAndSymbolNames):
    gates, symbolNamesDict = getGatesAndSymbolNames
    for gate in gates:
        if gate.exchangeName.lower() == "binance":
            continue
        symbolName = symbolNamesDict[gate.exchangeName]
        try:
            verifiedOrder = gate.create_and_test_spot_order(
                symbolName,
                "BUY",
                "LIMIT",
                quantity=0.02,
                price=30000,
                time_in_force="GTC",
                new_client_order_id=str(int(time.time())),
            )
            result = gate.make_spot_order(verifiedOrder)
        except Exception as e:
            assert False, "Problem in making order from {} exchange: {}".format(
                gate.exchangeName, str(e)
            )

        # print('Submitted order on {} exchange: {}'.format(gate.exchangeName, result))

        order = gate.get_order(symbolName, order_id=result["orderId"])
        assert (
            order["clientOrderId"] == result["clientOrderId"]
        ), "Fetch client orderID is not equal to the actual client orderID from {} exchange.".format(
            gate.exchangeName
        )
        # print('Correct \'clientOrderId\'.')

        order = gate.get_order(symbolName, local_order_id=result["clientOrderId"])
        assert (
            order["orderId"] == result["orderId"]
        ), "Fetch orderID is not equal to the actual orderID from {} exchange.".format(
            gate.exchangeName
        )
        # print('Correct \'orderId\'.')

        gate.cancel_order("BTCUSDT", order_id=result["orderId"])


def testCancelingAllOpenOrders(getGatesAndSymbolNames):
    gates, symbolNamesDict = getGatesAndSymbolNames
    for gate in gates:
        symbolName = symbolNamesDict[gate.exchangeName]
        result = gate.cancel_all_symbol_open_orders(symbolName)

        openOrders = gate.get_open_orders(symbolName)
        assert (
            len(openOrders) == 0
        ), "Problem in canceling all Open Orders in {} exchange.".format(
            gate.exchangeName
        )


def testCancelingOrder(getGatesAndSymbolNames):
    gates, symbolNamesDict = getGatesAndSymbolNames
    for gate in gates:
        symbolName = symbolNamesDict[gate.exchangeName]
        try:
            verifiedOrder = gate.create_and_test_spot_order(
                symbolName,
                "BUY",
                "LIMIT",
                quantity=0.002,
                price=28000,
                time_in_force="GTC",
                new_client_order_id=str(int(time.time())),
            )
            result = gate.make_spot_order(verifiedOrder)
        except Exception as e:
            assert False, "Problem in making order in {} exchange: {}".format(
                gate.exchangeName, str(e)
            )

        result = gate.cancel_order(symbol=symbolName, order_id=result["orderId"])
        result = gate.get_order(symbol=symbolName, order_id=result["orderId"])

        assert result["status"].upper() in [
            "CANCELED",
            "CANCELLED",
        ], "Problem in canceling specified Open Orders in {} exchange.".format(
            gate.exchangeName
        )

        try:
            verifiedOrder = gate.create_and_test_spot_order(
                symbolName,
                "BUY",
                "LIMIT",
                quantity=0.002,
                price=28000,
                time_in_force="GTC",
                new_client_order_id=str(int(time.time())),
            )
            result = gate.make_spot_order(verifiedOrder)
        except Exception as e:
            assert False, "Problem in making order in {} exchange: {}".format(
                gate.exchangeName, str(e)
            )

        result = gate.cancel_order(
            symbol=symbolName, local_order_id=result["clientOrderId"]
        )
        result = gate.get_order(symbol=symbolName, order_id=result["orderId"])

        assert result["status"].upper() in [
            "CANCELED",
            "CANCELLED",
        ], "Problem in canceling specified Open Orders in {} exchange.".format(
            gate.exchangeName
        )



from kucoin.websocket.websocket import ConnectWebsocket
import asyncio


class KucoinWsClient:
    """
    https://docs.kucoin.com/#websocket-feed
    """

    def __init__(self):
        self._callback = None
        self._conn = None
        self._loop = None
        self._client = None
        self._private = False
        self._topics = set()

    @classmethod
    async def create(cls, loop, client, callback, private=False):
        self = KucoinWsClient()
        loop = loop if loop else asyncio.get_running_loop()
        self._loop = loop
        self._client = client
        self._private = private
        self._callback = callback
        self._conn = ConnectWebsocket(loop, self._client, self._recv, private)
        return self

    async def _recv(self, msg):
        if 'data' in msg:
            await self._callback(msg)

    async def subscribe(self, topic):
        """Subscribe to a channel
        :param topic: required
        :type topic: str
        :returns: None
        """

        req_msg = {
            'type': 'subscribe',
            'topic': topic,
            'response': True
        }
        self._conn.topics.append(topic)
        await self._conn.send_message(req_msg)

    async def unsubscribe(self, topic):
        """Unsubscribe from a topic

        :param topic: required
        :type topic: str
        :returns: None
        """

        req_msg = {
            'type': 'unsubscribe',
            'topic': topic,
            'response': True
        }
        self._conn.topics.remove(topic)
        await self._conn.send_message(req_msg)



import requests
from binance_f.exception.binanceapiexception import BinanceApiException
from binance_f.impl.utils import *
# from binance_f.base.printobject import *


def check_response(json_wrapper):
    if json_wrapper.contain_key("success"):
        success = json_wrapper.get_boolean("success")
        if success is False:
            err_code = json_wrapper.get_int_or_default("code", "")
            err_msg = json_wrapper.get_string_or_default("msg", "")
            if err_code == "":
                raise BinanceApiException(BinanceApiException.EXEC_ERROR, "[Executing] " + err_msg)
            else:
                raise BinanceApiException(BinanceApiException.EXEC_ERROR, "[Executing] " + str(err_code) + ": " + err_msg)
    elif json_wrapper.contain_key("code"):
        code = json_wrapper.get_int("code")
        msg = json_wrapper.get_string_or_default("msg", "")
        if code != 200:
            raise BinanceApiException(BinanceApiException.EXEC_ERROR, "[Executing] " + str(code) + ": " + msg)

def get_limits_usage(response):
    limits = {}
    limits_headers = ["X-MBX-USED-WEIGHT-", "X-MBX-ORDER-COUNT-" ]  # Limit headers to catch
    for key,value in response.headers.items():
        if any([key.startswith(h) for h in limits_headers]):
            limits[key] = value
    return limits

def call_sync(request):
    if request.method == "GET":
        response = requests.get(request.host + request.url, headers=request.header)
        limits = get_limits_usage(response)
        json_wrapper = parse_json_from_string(response.text)
        check_response(json_wrapper)
        return (request.json_parser(json_wrapper),limits)
    elif request.method == "POST":
        response = requests.post(request.host + request.url, headers=request.header)
        limits = get_limits_usage(response)
        json_wrapper = parse_json_from_string(response.text)
        check_response(json_wrapper)
        return (request.json_parser(json_wrapper),limits)
    elif request.method == "DELETE":
        response = requests.delete(request.host + request.url, headers=request.header)
        limits = get_limits_usage(response)
        json_wrapper = parse_json_from_string(response.text)
        check_response(json_wrapper)
        return (request.json_parser(json_wrapper),limits)
    elif request.method == "PUT":
        response = requests.put(request.host + request.url, headers=request.header)
        limits = get_limits_usage(response)
        json_wrapper = parse_json_from_string(response.text)
        check_response(json_wrapper)
        return (request.json_parser(json_wrapper),limits)





class Balance:

    def __init__(self):
        self.asset = ""
        self.walletBalance = 0.0
        self.crossWallet = 0.0

    @staticmethod
    def json_parse(json_data):
        result = Balance()
        result.asset = json_data.get_string("a")
        result.walletBalance = json_data.get_float("wb")
        result.crossWallet = json_data.get_float("cw")
        return result


class Position:

    def __init__(self):
        self.symbol = ""
        self.amount = 0.0
        self.entryPrice = 0.0
        self.preFee = 0.0
        self.unrealizedPnl = 0.0
        self.marginType = ""
        self.isolatedWallet = 0.0
        self.positionSide = ""

    @staticmethod
    def json_parse(json_data):
        result = Position()
        result.symbol = json_data.get_string("s")
        result.amount = json_data.get_float("pa")
        result.entryPrice = json_data.get_float("ep")
        result.preFee = json_data.get_float("cr")
        result.unrealizedPnl = json_data.get_float("up")
        result.marginType = json_data.get_string("mt")
        result.isolatedWallet = json_data.get_float("iw")
        result.positionSide = json_data.get_string("ps")
        return result


class AccountUpdate:
    def __init__(self):
        self.eventType = ""
        self.eventTime = 0
        self.transactionTime = 0
        self.balances = list()
        self.positions = list()

    @staticmethod
    def json_parse(json_data):
        result = AccountUpdate()
        result.eventType = json_data.get_string("e")
        result.eventTime = json_data.get_int("E")
        result.transactionTime = json_data.get_int("T")

        data_group = json_data.get_object("a")
        
        element_list = list()
        data_list = data_group.get_array("B")
        for item in data_list.get_items():
            element = Balance.json_parse(item)
            element_list.append(element)
        result.balances = element_list
       
        if data_group.contain_key("P"):
            element_list = list()
            data_list = data_group.get_array("P")
            for item in data_list.get_items():
                element = Position.json_parse(item)
                element_list.append(element)
            result.positions = element_list
        return result



import time


class OrderData:
    def __init__(self, symbol, side, order_type):
        self.symbol = symbol
        self.side = side
        self.order_type = order_type

        self.time_in_force = None
        self.quantity = None
        self.quote_order_qty = None
        self.price = None
        self.new_client_order_id = None
        self.stop_price = None
        self.iceberg_qty = None
        self.new_order_resp_type = None
        self.timestamp = None
        self.extra_params = None
        self.recv_window = None

    def set_time_in_force(self, time_in_force):
        self.time_in_force = time_in_force

    def set_quantity(self, quantity):
        self.quantity = quantity

    def set_quote_order_qty(self, quote_order_qty):
        self.quote_order_qty = quote_order_qty

    def set_price(self, price):
        self.price = price

    def set_new_client_order_id(self, new_client_order_id):
        self.new_client_order_id = new_client_order_id

    def set_stop_price(self, stop_price):
        self.stop_price = stop_price

    def set_iceberg_qty(self, iceberg_qty):
        self.iceberg_qty = iceberg_qty

    def set_new_order_resp_type(self, new_order_resp_type):
        self.new_order_resp_type = new_order_resp_type

    def set_timestamp(self):
        self.timestamp = time.time()

    def set_extra_params(self, extra_params):
        self.extra_params = extra_params


class FuturesOrderData:
    def __init__(self, symbol, side=None, order_type=None):
        self.symbol = symbol
        self.side = side
        self.orderType = order_type

        self.position_side = None
        self.time_in_force = None
        self.quantity = None
        self.quote_quantity = None
        self.reduce_only = None
        self.price = None
        self.new_client_order_id = None
        self.stop_price = None
        self.close_position = None
        self.activation_price = None
        self.callback_rate = None
        self.working_type = None
        self.price_protect = None
        self.new_order_resp_type = None
        self.extra_params = None
        self.leverage = None

    def set_order_side(self, order_side):
        self.side = order_side

    def set_position_side(self, position_side):
        self.position_side = position_side

    def set_time_in_force(self, time_in_force):
        self.time_in_force = time_in_force

    def set_quantity(self, quantity):
        self.quantity = quantity

    def set_reduce_only(self, reduce_only):
        self.reduce_only = reduce_only

    def set_price(self, price):
        self.price = price

    def set_new_client_order_id(self, new_client_order_id):
        self.new_client_order_id = new_client_order_id

    def set_stop_price(self, stop_price):
        self.stop_price = stop_price

    def set_close_position(self, close_position):
        self.close_position = close_position

    def set_activation_price(self, activation_price):
        self.activation_price = activation_price

    def set_callback_rate(self, callback_rate):
        self.callback_rate = callback_rate

    def set_working_type(self, working_type):
        self.working_type = working_type

    def set_price_protect(self, price_protect):
        self.price_protect = price_protect

    def set_new_order_resp_type(self, new_order_resp_type):
        self.new_order_resp_type = new_order_resp_type

    def set_extra_params(self, extra_params):
        self.extra_params = extra_params

    def set_leverage(self, leverage):
        self.leverage = leverage

    def set_quote_quantity(self, quote_quantity):
        self.quote_quantity = quote_quantity


def set_spot_order_data(
    iceberg_qty,
    new_client_order_id,
    new_order_resp_type,
    order_type,
    price,
    quantity,
    quote_order_qty,
    side,
    stop_price,
    symbol,
    time_in_force,
    extra_params=None,
):
    curr_order = OrderData(symbol.upper(), side.upper(), order_type.upper())
    if quantity is not None:
        curr_order.set_quantity(quantity)
    if quote_order_qty is not None:
        curr_order.set_quote_order_qty(quote_order_qty)
    if price is not None:
        curr_order.set_price(price)
    if time_in_force is not None:
        curr_order.set_time_in_force(time_in_force)
    if stop_price is not None:
        curr_order.set_stop_price(stop_price)
    if iceberg_qty is not None:
        curr_order.set_iceberg_qty(iceberg_qty)
    if new_order_resp_type is not None:
        curr_order.set_new_order_resp_type(new_order_resp_type)
    if new_client_order_id is not None:
        curr_order.set_new_client_order_id(new_client_order_id)
    if extra_params is not None:
        curr_order.set_extra_params(extra_params)
    return curr_order


def set_futures_order_data(
    activation_price,
    callback_rate,
    close_position,
    extra_params,
    new_client_order_id,
    new_order_resp_type,
    order_type,
    position_side,
    price,
    price_protect,
    quantity,
    reduce_only,
    side,
    stop_price,
    symbol,
    time_in_force,
    working_type,
    quote_quantity,
):
    if extra_params is None:
        extra_params = {}
    curr_order = FuturesOrderData(symbol=symbol.upper(), order_type=order_type.upper())
    if side is not None:
        curr_order.set_order_side(side)
    if position_side is not None:
        curr_order.set_position_side(position_side)
    if time_in_force is not None:
        curr_order.set_time_in_force(time_in_force)
    if quantity is not None:
        curr_order.set_quantity(quantity)
    if quote_quantity is not None:
        curr_order.set_quote_quantity(quote_quantity)
    if reduce_only is not None:
        curr_order.set_reduce_only(reduce_only)
    if price is not None:
        curr_order.set_price(price)
    if new_client_order_id is not None:
        curr_order.set_new_client_order_id(new_client_order_id)
    if stop_price is not None:
        curr_order.set_stop_price(stop_price)
    if close_position is not None:
        curr_order.set_close_position(close_position)
    if activation_price is not None:
        curr_order.set_activation_price(activation_price)
    if callback_rate is not None:
        curr_order.set_callback_rate(callback_rate)
    if working_type is not None:
        curr_order.set_working_type(working_type)
    if price_protect is not None:
        curr_order.set_price_protect(price_protect)
    if new_order_resp_type is not None:
        curr_order.set_new_order_resp_type(new_order_resp_type)
    if "leverage" in extra_params.keys():
        curr_order.set_leverage(extra_params["leverage"])
    if extra_params is not None:
        curr_order.set_extra_params(extra_params)
    return curr_order


def get_quantity(enter_price, quantity, quote_quantity, step_precision):
    if (quantity is not None and quote_quantity is not None) or (
        quantity is None and quote_quantity is None
    ):
        raise ValueError("Specify either quantity or quoteQuantity and not both")
    if quantity is None:
        if float(step_precision) > 0.5:
            quantity = round(
                quote_quantity / enter_price, len(str(float(step_precision))) - 3
            )
        else:
            quantity = round(
                quote_quantity / enter_price, len(str(float(step_precision))) - 2
            )
    return quantity


def get_tp_sl_limit_order_ids(ordering_result):
    order_ids = {}
    for order in ordering_result:
        if order["type"] == "LIMIT":
            order_ids["mainOrder"] = order["orderId"]
        elif order["type"] == "STOP_MARKET":
            order_ids["stopLoss"] = order["orderId"]
        elif order["type"] == "TAKE_PROFIT_MARKET":
            order_ids["takeProfit"] = order["orderId"]
    return order_ids


def get_tp_sl_market_order_ids(ordering_result, has_sl, has_tp):
    order_ids = {}
    for order in ordering_result:
        if order["type"] == "MARKET":
            order_ids["mainOrder"] = order["orderId"]
        elif order["type"] == "STOP_MARKET":
            order_ids["stopLoss"] = order["orderId"]
        elif order["type"] == "TAKE_PROFIT_MARKET":
            order_ids["takeProfit"] = order["orderId"]

    if "mainOrder" not in order_ids.keys():
        raise RuntimeError("Problem in main order")

    if has_sl and "stopLoss" not in order_ids.keys():
        raise RuntimeError("Problem in stop loss order")

    if has_tp and "takeProfit" not in order_ids.keys():
        raise RuntimeError("Problem in take profit order")
    return order_ids



#!/usr/bin/python
# -*- coding:utf-8 -*-

import json
import requests
import hmac
import hashlib
import base64
import time
from uuid import uuid1
from urllib.parse import urljoin

try:
    import pkg_resources
    version = 'v' + pkg_resources.get_distribution("kucoin-futures-python").version
except (ModuleNotFoundError, pkg_resources.DistributionNotFound):
    version = 'v1.0.0'
    
class KucoinFuturesBaseRestApi(object):

    def __init__(self, key='', secret='', passphrase='', is_sandbox=False, url='', is_v1api=False):
        """
        https://docs.kumex.com

        :param key: Api Token Id  (Mandatory)
        :type key: string
        :param secret: Api Secret  (Mandatory)
        :type secret: string
        :param passphrase: Api Passphrase used to create API  (Mandatory)
        :type passphrase: string
        :param is_sandbox: True sandbox , False  (optional)
        """
        if url:
            self.url = url
        else:
            if is_sandbox:
               self.url = 'https://api-sandbox-futures.kucoin.com'
            else:
                self.url = 'https://api-futures.kucoin.com'
        self.key = key
        self.secret = secret
        self.passphrase = passphrase
        self.is_v1api = is_v1api

    def _request(self, method, uri, timeout=5, auth=True, params=None):
        uri_path = uri
        data_json = ''
        if method in ['GET', 'DELETE']:
            if params:
                strl = []
                for key in sorted(params):
                    strl.append("{}={}".format(key, params[key]))
                data_json += '&'.join(strl)
                uri += '?' + data_json
                uri_path = uri
        else:
            if params:
                data_json= json.dumps(params)

                uri_path = uri + data_json

        headers = {}
        if auth:
            now_time = int(time.time()) * 1000
            str_to_sign = str(now_time) + method + uri_path
            sign = base64.b64encode(
                hmac.new(self.secret.encode('utf-8'), str_to_sign.encode('utf-8'), hashlib.sha256).digest())
            if self.is_v1api:
                headers = {
                    "KC-API-SIGN": sign,
                    "KC-API-TIMESTAMP": str(now_time),
                    "KC-API-KEY": self.key,
                    "KC-API-PASSPHRASE": self.passphrase,
                    "Content-Type": "application/json"
                }
            else:
                passphrase = base64.b64encode(
                    hmac.new(self.secret.encode('utf-8'), self.passphrase.encode('utf-8'), hashlib.sha256).digest())
                headers = {
                    "KC-API-SIGN": sign,
                    "KC-API-TIMESTAMP": str(now_time),
                    "KC-API-KEY": self.key,
                    "KC-API-PASSPHRASE": passphrase,
                    "Content-Type": "application/json",
                    "KC-API-KEY-VERSION": "2"
                }
        headers["User-Agent"] = "kucoin-futures-python-sdk/" + version
        url = urljoin(self.url, uri)

        if method in ['GET', 'DELETE']:
            response_data = requests.request(method, url, headers=headers, timeout=timeout)
        else:
            response_data = requests.request(method, url, headers=headers, data=data_json,
                                             timeout=timeout)
        return self.check_response_data(response_data)

    @staticmethod
    def check_response_data(response_data):
        if response_data.status_code == 200:
            try:
                data = response_data.json()
            except ValueError:
                raise Exception(response_data.content)
            else:
                if data and data.get('code'):
                    if data.get('code') == '200000':
                        if data.get('data'):
                            return data['data']
                        else:
                            return data
                    else:
                        raise Exception("{}-{}".format(response_data.status_code, response_data.text))
        else:
            raise Exception("{}-{}".format(response_data.status_code, response_data.text))

    @property
    def return_unique_id(self):
        return ''.join([each for each in str(uuid1()).split('-')])



class Order:

    def __init__(self):
        self.price = 0.0
        self.qty = 0.0


class OrderBookEvent:

    def __init__(self):
        self.eventType = ""
        self.eventTime = ""
        self.transactionTime = ""
        self.symbol = ""
        self.firstUpdateId = 0
        self.lastUpdateId = 0
        self.lastUpdateIdInlastStream = 0
        self.bids = list()
        self.asks = list()

    @staticmethod
    def json_parse(json_data):
        result = OrderBookEvent()
        result.eventType = json_data.get_string("e")
        result.eventTime = json_data.get_int("E")
        result.transactionTime = json_data.get_int("T")
        result.symbol = json_data.get_string("s")
        result.firstUpdateId = json_data.get_int("U")
        result.lastUpdateId = json_data.get_int("u")
        result.lastUpdateIdInlastStream = json_data.get_int("pu")

        list_array = json_data.get_array("b")
        bid_list = list()
        for item in list_array.get_items():
            order = Order()
            val = item.convert_2_list()
            order.price = val[0]
            order.qty = val[1]
            bid_list.append(order)
        result.bids = bid_list

        list_array = json_data.get_array("a")
        ask_list = list()
        for item in list_array.get_items():
            order = Order()
            val = item.convert_2_list()
            order.price = val[0]
            order.qty = val[1]
            ask_list.append(order)
        result.asks = ask_list        

        return result


class Composition:

    def __init__(self):
        self.baseAsset = ""
        self.wQty = 0.0  # Weight in quantity
        self.wPct = 0.0  # Weight in percentage

    @staticmethod
    def json_parse(json_data):
        data_obj = Composition()
        data_obj.baseAsset = json_data.get_string("b")
        data_obj.wQty = json_data.get_float("w")
        data_obj.wPct = json_data.get_float("W")

class CompositeIndexEvent:

    def __init__(self):
        self.eventType = ""
        self.eventTime = 0
        self.symbol = ""
        self.price = 0.0
        self.composition = list()

    @staticmethod
    def json_parse(json_data):
        composite_index_event = CompositeIndexEvent()
        composite_index_event.eventType = json_data.get_string("e")
        composite_index_event.eventTime = json_data.get_int("E")
        composite_index_event.symbol = json_data.get_string("s")
        composite_index_event.price = json_data.get_folat("p")

        list_array = json_data.get_array("c")
        composition_list = list()
        for item in list_array.get_items():
            c = Composition.json_parse(item)
            composition_list.append(c)
        composite_index_event.composition = composition_list

        return composite_index_event


import warnings

from kucoin.base_request.base_request import KucoinBaseRestApi


class MarketData(KucoinBaseRestApi):

    def get_symbol_list(self, **kwargs):
        """
        https://docs.kucoin.com/#get-symbols-list
        :param kwargs: [Optional] market
        :return:
        [
          {
            "symbol": "BTC-USDT",
            "name": "BTC-USDT",
            "baseCurrency": "BTC",
            "quoteCurrency": "USDT",
            "baseMinSize": "0.00000001",
            "quoteMinSize": "0.01",
            "baseMaxSize": "10000",
            "quoteMaxSize": "100000",
            "baseIncrement": "0.00000001",
            "quoteIncrement": "0.01",
            "priceIncrement": "0.00000001",
            "feeCurrency": "USDT",
            "enableTrading": true,
            "isMarginEnabled": true,
            "priceLimitRate": "0.1"
          }
        ]
        """
        params = {}
        if kwargs:
            params.update(kwargs)
        return self._request('GET', '/api/v1/symbols', params=params)

    def get_ticker(self, symbol):
        """
        https://docs.kucoin.com/#get-ticker
        :param symbol: symbol (Mandatory)
        :type: str
        :return:
        {
            "sequence": "1550467636704",
            "bestAsk": "0.03715004",
            "size": "0.17",
            "price": "0.03715005",
            "bestBidSize": "3.803",
            "bestBid": "0.03710768",
            "bestAskSize": "1.788",
            "time": 1550653727731

        }
        """
        params = {
            'symbol': symbol
        }
        return self._request('GET', '/api/v1/market/orderbook/level1', params=params)

    def get_all_tickers(self):
        """
        https://docs.kucoin.com/#get-all-tickers
        :return:
        {
            "time": 1550653727731,
            "ticker": [
              {
                "symbol": "BTC-USDT",
                "symbolName": "BTC-USDT",
                "buy": "0.00001191",
                "sell": "0.00001206",
                "changeRate": "0.057",
                "changePrice": "0.00000065",
                "high": "0.0000123",
                "low": "0.00001109",
                "vol": "45161.5073",
                "volValue": "2127.28693026”,
                "last": "0.00001204"
              },
              {
                "symbol": "BCD-BTC",
                "symbolName": "BCD-BTC",
                "buy": "0.00018564",
                "sell": "0.0002",
                "changeRate": "-0.0753",
                "changePrice": "-0.00001522",
                "high": "0.00021489",
                "low": "0.00018351",
                "vol": "72.99679763",
                "volValue": "2127.28693026”,
                "last": "0.00018664"
              }
            ]
        }
        """
        return self._request('GET', '/api/v1/market/allTickers')

    def get_24h_stats(self, symbol):
        """
        https://docs.kucoin.com/#get-24hr-stats
        :param symbol: symbol (Mandatory)
        :type: str
        :return:
        {
            "symbol": "ETH-BTC",    // symbol
            "high": "0.03736329",   // 24h highest price
            "vol": "2127.286930263025",  // 24h volume，the aggregated trading volume in ETH
            "volValue": "43.58567564",  // 24h total, the trading volume in quote currency of last 24 hours
            "last": "0.03713983",   // last price
            "low": "0.03651252",    // 24h lowest price
            "buy": "0.03712118",    // bestAsk
            "sell": "0.03713983",   // bestBid
            "changePrice": "0.00037224",  // 24h change price
            "averagePrice": "8699.24180977",//24h average transaction price yesterday
            "time": 1550847784668,  //time
            "changeRate": "0.0101" // 24h change rate
        }
        """
        params = {
            'symbol': symbol
        }
        return self._request('GET', '/api/v1/market/stats', params=params)

    def get_market_list(self):
        """
        https://docs.kucoin.com/#get-market-list
        :return:
        [
            "BTC",
            "KCS",
            "USDS",  //SC has been changed to USDS
            "ALTS" //ALTS market includes ETH, NEO, TRX
        ]
        """
        return self._request('GET', '/api/v1/markets')

    def get_part_order(self, pieces, symbol):
        """
        https://docs.kucoin.com/#get-part-order-book-aggregated
        :param pieces: pieces of data (ask and bid data) on the order book (Mandatory)
        :type: int
        :param symbol: symbol
        :type: str
        :return:
        {
            "sequence": "3262786978",
            "time": 1550653727731,
            "bids": [["6500.12", "0.45054140"],
                     ["6500.11", "0.45054140"]],  //[price，size]
            "asks": [["6500.16", "0.57753524"],
                     ["6500.15", "0.57753524"]]
        }
        """
        params = {
            'symbol': symbol
        }
        return self._request('GET', '/api/v1/market/orderbook/level2_{pieces}'.format(pieces=pieces), params=params)

    def get_aggregated_orderv3(self, symbol):
        """
        https://docs.kucoin.com/#get-full-order-book-aggregated
        :param symbol: symbol (Mandatory)
        :type: str
        :return:
        {
            "sequence": "3262786978",
            "time": 1550653727731,
            "bids": [["6500.12", "0.45054140"],
                     ["6500.11", "0.45054140"]],  //[price，size]
            "asks": [["6500.16", "0.57753524"],
                     ["6500.15", "0.57753524"]]
        }
        """
        params = {
            'symbol': symbol
        }
        return self._request('GET', '/api/v3/market/orderbook/level2', params=params)

    def get_aggregated_order(self, symbol):
        """
        https://docs.kucoin.com/#get-full-order-book-aggregated
        :param symbol: symbol (Mandatory)
        :type: str
        :return:
        {
            "sequence": "3262786978",
            "time": 1550653727731,
            "bids": [["6500.12", "0.45054140"],
                     ["6500.11", "0.45054140"]],  //[price，size]
            "asks": [["6500.16", "0.57753524"],
                     ["6500.15", "0.57753524"]]
        }
        """
        params = {
            'symbol': symbol
        }
        return self._request('GET', '/api/v3/market/orderbook/level2', params=params)

    def get_atomic_orderv3(self, symbol):
        """
        https://docs.kucoin.com/#get-full-order-book-atomic
        :param symbol: symbol (Mandatory)
        :type: str
        :return:
        {
            "data": {
                "sequence": "1573503933086",
                "asks": [
                    [
                        "5e0d672c1f311300093ac522",   //orderId
                        "0.1917",                     //price
                        "390.9275",                   //size
                        "1577936689346546088"         //time,nanoseconds
                    ],
                    [
                        "5e0d672891432f000819ecc3",
                        "0.19171",
                        "1456.1316",
                        "1577936685718811031"
                    ]
                ],
                "bids": [
                    [
                        "5e0d672cdc53860007f30262",
                        "0.19166",
                        "178.1936",
                        "1577936689166023452"
                    ],
                    [
                        "5e0d671a91432f000819d1b0",
                        "0.19165",
                        "583.6298",
                        "1577936671595901518"
                    ]
                ],
                "time": 1577936685107
            }
        }
        """
        return self._request('GET', f'/api/v3/market/orderbook/level3?symbol={symbol}')

    def get_atomic_order(self, symbol):
        """
        https://docs.kucoin.com/#get-full-order-book-atomic
        :param symbol: symbol (Mandatory)
        :type: str
        :return:
        {
            "data": {
                "sequence": "1573503933086",
                "asks": [
                    [
                        "5e0d672c1f311300093ac522",   //orderId
                        "0.1917",                     //price
                        "390.9275",                   //size
                        "1577936689346546088"         //time,nanoseconds
                    ],
                    [
                        "5e0d672891432f000819ecc3",
                        "0.19171",
                        "1456.1316",
                        "1577936685718811031"
                    ]
                ],
                "bids": [
                    [
                        "5e0d672cdc53860007f30262",
                        "0.19166",
                        "178.1936",
                        "1577936689166023452"
                    ],
                    [
                        "5e0d671a91432f000819d1b0",
                        "0.19165",
                        "583.6298",
                        "1577936671595901518"
                    ]
                ],
                "time": 1577936685107
            }
        }
        """
        return self._request('GET', f'/api/v1/market/orderbook/level3?symbol={symbol}')

    def get_trade_histories(self, symbol):
        """
        https://docs.kucoin.com/#get-trade-histories
        :param symbol: symbol (Mandatory)
        :type: str
        :return:
        [
          {
              "sequence": "1545896668571",
              "price": "0.07",                      //Filled price
              "size": "0.004",                      //Filled amount
              "side": "buy",                        //Filled side. The filled side is set to the taker by default.
              "time": 1545904567062140823           //Transaction time
          },
          {
              "sequence": "1545896668578",
              "price": "0.054",
              "size": "0.066",
              "side": "buy",
              "time": 1545904581619888405
          }
        ]
        """
        params = {
            'symbol': symbol
        }
        return self._request('GET', '/api/v1/market/histories', params=params)

    def get_kline(self, symbol, kline_type, **kwargs):
        """
        https://docs.kucoin.com/#get-klines
        :param symbol: symbol (Mandatory)
        :type: str
        :param kline_type: Type of candlestick patterns (Mandatory)
        :type: str
        :param kwargs: [Optional] startAt, endAt, currentPage, pageSize
        :return:
        [
          [
              "1545904980",             //Start time of the candle cycle
              "0.058",                  //opening price
              "0.049",                  //closing price
              "0.058",                  //highest price
              "0.049",                  //lowest price
              "0.018",                  //Transaction amount
              "0.000945"                //Transaction volume
          ],
          [
              "1545904920",
              "0.058",
              "0.072",
              "0.072",
              "0.058",
              "0.103",
              "0.006986"
          ]
        ]
        """
        params = {
            'symbol': symbol,
            'type': kline_type
        }
        if kwargs:
            params.update(kwargs)
        return self._request('GET', '/api/v1/market/candles', params=params)

    def get_currencies(self):
        """
        https://docs.kucoin.com/#get-currencies
        :return:
        [{
            "currency": "BTC",
            "name": "BTC",
            "fullName": "Bitcoin",
            "precision": 8,
            "withdrawalMinSize": "0.002",
            "withdrawalMinFee": "0.0005",
            "isWithdrawEnabled": true,
            "isDepositEnabled": true,
            "isMarginEnabled": true,
            "isDebitEnabled": true
        },
        {

            "currency": "ETH",
            "name": "ETH",
            "fullName": "Ethereum",
            "precision": 8,
            "withdrawalMinSize": "0.02",
            "withdrawalMinFee": "0.01",
            "isWithdrawEnabled": true,
            "isDepositEnabled": true,
            "isMarginEnabled": true,
            "isDebitEnabled": true

        }]
        """
        return self._request('GET', '/api/v1/currencies')

    def get_currency_detail(self, currency, chain=None):
        """
        https://docs.kucoin.com/#get-currency-detail
        :param currency: currency (Mandatory)
        :type: str
        :param chain: [Optional] Support for querying the chain of currency, e.g. The available value for USDT are OMNI,
        ERC20, TRC20. This only apply for multi-chain currency, and there is no need for single chain currency.
        :type: str
        :return:
        {
            "currency": "BTC",
            "name": "BTC",
            "fullName": "Bitcoin",
            "precision": 8,
            "withdrawalMinSize": "0.002",
            "withdrawalMinFee": "0.0005",
            "isWithdrawEnabled": true,
            "isDepositEnabled": true,
            "isMarginEnabled": true,
            "isDebitEnabled": true
        }
        """
        warnings.warn("The 'get_currency_detail' method is deprecated, use 'get_currency_detail_v2' instead",
                      DeprecationWarning)
        params = {}
        if chain:
            params['chain'] = chain
        return self._request('GET', '/api/v1/currencies/{currency}'.format(currency=currency), params=params)

    def get_currency_detail_v2(self, currency, chain=None):
        """
        https://docs.kucoin.com/#get-currency-detail-recommend
        :param currency: currency (Mandatory)
        :type: str
        :param chain: [Optional] Support for querying the chain of currency, return the currency details of all chains by default.
        :type: str
        :return:
        {
          "currency": "BTC",
          "name": "BTC",
          "fullName": "Bitcoin",
          "precision": 8,
          "confirms": null,
          "contractAddress": null,
          "isMarginEnabled": true,
          "isDebitEnabled": true,
          "chains": [
            {
              "chainName": "BTC",
              "withdrawalMinSize": "0.0008",
              "withdrawalMinFee": "0.0005",
              "isWithdrawEnabled": true,
              "isDepositEnabled": true,
              "confirms": 2,
              "contractAddress": ""
            },
            {
              "chainName": "KCC",
              "withdrawalMinSize": "0.0008",
              "withdrawalMinFee": "0.00002",
              "isWithdrawEnabled": true,
              "isDepositEnabled": true,
              "confirms": 20,
              "contractAddress": ""
            },
            {
              "chainName": "TRC20",
              "withdrawalMinSize": "0.0008",
              "withdrawalMinFee": "0.0004",
              "isWithdrawEnabled": false,
              "isDepositEnabled": true,
              "confirms": 1,
              "contractAddress": ""
            },
            {
              "chainName": "BTC-Segwit",
              "withdrawalMinSize": "0.0008",
              "withdrawalMinFee": "0.0005",
              "isWithdrawEnabled": true,
              "isDepositEnabled": true,
              "confirms": 2,
              "contractAddress": ""
            }
          ]
        }
        """
        params = {}
        if chain:
            params['chain'] = chain
        return self._request('GET', '/api/v2/currencies/{currency}'.format(currency=currency), params=params)

    def get_fiat_price(self, **kwargs):
        """
        https://docs.kucoin.com/#get-fiat-price
        :param kwargs: [Optional] base, currencies
        :return:
        {
            "BTC": "3911.28000000",
            "ETH": "144.55492453",
            "LTC": "48.45888179",
            "KCS": "0.45546856"
        }
        """
        params = {}
        if kwargs:
            params.update(kwargs)
        return self._request('GET', '/api/v1/prices', params=params)

    def get_server_timestamp(self):
        """
        https://docs.kucoin.com/#server-time

        get server timestamp
        :return: 1570609496404
        """
        return self._request('GET', '/api/v1/timestamp', auth=False)

    def get_server_status(self):
        """
        https://docs.kucoin.com/#server-time

        get server timestamp
        :return:
        {
          "status": "open",                //open, close, cancelonly
          "msg":  "upgrade match engine"   //remark for operation
        }
        """
        return self._request('GET', '/api/v1/status', auth=False)



import json
import logging
import time

import pytest

from TradeGates.TradeGate import TradeGate

loglevel = logging.INFO
logging.basicConfig(level=loglevel)
log = logging.getLogger(__name__)


@pytest.fixture
def getGatesAndSymbolNames():
    gates = []
    symbolNames = {}
    with open('../../config.json') as f:
        config = json.load(f)

    for key in config.keys():
        gates.append(TradeGate(config[key], sandbox=True))
        if gates[-1].exchangeName.lower() == 'kucoin':
            symbolNames[gates[-1].exchangeName] = 'XBTUSDTM'
        else:
            symbolNames[gates[-1].exchangeName] = 'BTCUSDT'

    return gates, symbolNames


def testSymbolFuturesOrders(getGatesAndSymbolNames):
    gates, symbolNamesDict = getGatesAndSymbolNames
    for gate in gates:
        symbolName = symbolNamesDict[gate.exchangeName]
        symbolFutureOrders = gate.getSymbolOrders(symbolName, futures=True)
        # print('\nSymbol future orders from {} exchange: {}'.format(gate.exchangeName, symbolFutureOrders))
        assert symbolFutureOrders is not None, 'Problem in futures order list from {} exchange.'.format(
            gate.exchangeName)


def testFuturesBalance(getGatesAndSymbolNames):
    gates, symbolNamesDict = getGatesAndSymbolNames
    for gate in gates:
        balance = gate.getBalance(futures=True)
        # print('\nFutures balance from {} exchange: {}'.format(gate.exchangeName, balance))
        assert balance is not None, 'Problem in futures balance from {} exchange.'.format(gate.exchangeName)

        try:
            if gate.exchangeName != 'Binance':
                if sorted(list(balance[0].keys())) != sorted(['asset', 'free', 'locked', 'exchangeSpecific']):
                    assert False, 'Bad fetch balance interface for {} exchange,'.format(gate.exchangeName)
            else:
                if sorted(list(balance[0].keys())) != sorted(['asset', 'free', 'locked']):
                    assert False, 'Bad fetch balance interface for {} exchange,'.format(gate.exchangeName)
        except:
            assert False, 'Bad fetch single coin balance interface for {} exchange,'.format(gate.exchangeName)


def testFuturesSingleCoinBalance(getGatesAndSymbolNames):
    gates, symbolNamesDict = getGatesAndSymbolNames
    for gate in gates:
        balance = gate.getBalance('USDT', futures=True)
        # print('\nUSDT Futures balance from {} exchange: {}'.format(gate.exchangeName, balance))
        assert balance is not None, 'Problem in fetching futures single coin balance from {} exchange.'.format(
            gate.exchangeName)

        try:
            if gate.exchangeName != 'Binance':
                if sorted(list(balance.keys())) != sorted(['asset', 'free', 'locked', 'exchangeSpecific']):
                    assert False, 'Bad fetch balance interface for {} exchange,'.format(gate.exchangeName)
            else:
                if sorted(list(balance.keys())) != sorted(['asset', 'free', 'locked']):
                    assert False, 'Bad fetch balance interface for {} exchange,'.format(gate.exchangeName)
        except:
            assert False, 'Bad fetch single coin balance interface for {} exchange,'.format(gate.exchangeName)


def testCreatingFuturesOrder(getGatesAndSymbolNames):
    gates, symbolNamesDict = getGatesAndSymbolNames
    for gate in gates:
        symbolName = symbolNamesDict[gate.exchangeName]
        if gate.exchangeName.lower() == 'kucoin':
            extraParams = {'leverage': 5}
            futuresOrderData = gate.createAndTestFuturesOrder(symbolName, 'BUY', 'MARKET', quantity=0.002,
                                                              extraParams=extraParams)
        else:
            futuresOrderData = gate.createAndTestFuturesOrder(symbolName, 'BUY', 'MARKET', quantity=0.002)
        # print('\nTest creating futures order in {} exchange: {}'.format(gate.exchangeName, FuturesOrderData))
        assert futuresOrderData is not None, 'Problem in creating futures order in {} exchange.'.format(
            gate.exchangeName)


def testFuturesOrder(getGatesAndSymbolNames):
    gates, symbolNamesDict = getGatesAndSymbolNames
    for gate in gates:
        symbolName = symbolNamesDict[gate.exchangeName]
        if gate.exchangeName.lower() == 'kucoin':
            extraParams = {'leverage': 5}
            futuresOrderData = gate.createAndTestFuturesOrder(symbolName, 'BUY', 'MARKET', quantity=0.002,
                                                              extraParams=extraParams)
        else:
            futuresOrderData = gate.createAndTestFuturesOrder(symbolName, 'BUY', 'LIMIT', timeInForce='GTC',
                                                              price=20000, quantity=0.002)
        result = gate.makeFuturesOrder(futuresOrderData)
        print('\nFuture ordering in {} exchange: {}'.format(gate.exchangeName, result))
        assert result is not None, 'Problem in submitting futures order in {} exchange.'.format(gate.exchangeName)


def testBatchFuturesOrders(getGatesAndSymbolNames):
    gates, symbolNamesDict = getGatesAndSymbolNames
    for gate in gates:
        if gate.exchangeName.lower() == 'kucoin':
            continue
        symbolName = symbolNamesDict[gate.exchangeName]
        try:
            verifiedOrders = [gate.createAndTestFuturesOrder(symbolName, 'BUY', 'MARKET', quantity=0.1),
                              gate.createAndTestFuturesOrder(symbolName, 'BUY', 'MARKET', quantity=0.2),
                              gate.createAndTestFuturesOrder(symbolName, 'BUY', 'MARKET', quantity=0.3)]

            result = gate.makeBatchFuturesOrder(verifiedOrders)
            # print('\nResult of batch ordering from {} exchange: {}'.format(gate.exchangeName, result))
            assert result is not None, 'Problem in making new order in {} exchange'.format(gate.exchangeName)
        except Exception:
            assert False, 'Problem in making new order in {} exchange'.format(gate.exchangeName)


# @pytest.mark.skip(reason="For Special Purposes")
def testFuturesTpSlLimitOrder(getGatesAndSymbolNames):
    gates, symbolNamesDict = getGatesAndSymbolNames
    for gate in gates:
        symbolName = symbolNamesDict[gate.exchangeName]
        if gate.exchangeName.lower() != 'binance':
            continue
        try:
            result = gate.makeSlTpLimitFuturesOrder(symbol=symbolName, orderSide='BUY',
                                                    quantity=0.002, enterPrice=19000, takeProfit=20000,
                                                    stopLoss=18500, leverage=10, marginType='ISOLATED')
            print('\nResult of TP-SL-Limit ordering from {} exchange: {}'.format(gate.exchangeName, result))
            assert result is not None, 'Problem in making new order in {} exchange'.format(gate.exchangeName)
        except Exception:
            assert False, 'Problem in making new SL-TP-Limit order in {} exchange'.format(gate.exchangeName)


def testFuturesTpSlMarketOrder(getGatesAndSymbolNames):
    gates, symbolNamesDict = getGatesAndSymbolNames
    for gate in gates:
        symbolName = symbolNamesDict[gate.exchangeName]
        if gate.exchangeName.lower() != 'binance':
            continue
        try:
            result = gate.makeSlTpMarketFuturesOrder(symbol=symbolName, orderSide='BUY', quantity=None,
                                                     quoteQuantity=40,
                                                     takeProfit=20000, stopLoss=18500, leverage=10,
                                                     marginType='ISOLATED')
            print('\nResult of TP-SL-Market ordering from {} exchange: {}'.format(gate.exchangeName, result))
            assert result is not None, 'Problem in making new order in {} exchange'.format(gate.exchangeName)
        except Exception:
            assert False, 'Problem in making new SL-TP-Limit order in {} exchange'.format(gate.exchangeName)


def testGetFuturesOpenOrders(getGatesAndSymbolNames):
    gates, symbolNamesDict = getGatesAndSymbolNames
    for gate in gates:
        symbolName = symbolNamesDict[gate.exchangeName]
        symbolOpenOrders = gate.getOpenOrders(symbolName, futures=True)

        print('\n\'BTCUSDT\' open orders from {} exchange: {}'.format(gate.exchangeName, symbolOpenOrders))

        assert symbolOpenOrders is not None, \
            'Problem in getting list of open orders with symbol from {} exchange.'.format(gate.exchangeName)


def testGetPositionInformation(getGatesAndSymbolNames):
    gates, symbolNamesDict = getGatesAndSymbolNames
    for gate in gates:
        symbolName = symbolNamesDict[gate.exchangeName]
        openPosition = gate.getPositionInfo(symbolName)

        print('\nOpen position information from {} exchange: {}'.format(gate.exchangeName, openPosition))

        assert openPosition is not None, 'Problem in getting position information without symbol from {} exchange.'.format(
            gate.exchangeName)


def testGetFuturesOrder(getGatesAndSymbolNames):
    gates, symbolNamesDict = getGatesAndSymbolNames
    for gate in gates:
        if gate.exchangeName.lower() == 'bybit':
            continue
        symbolName = symbolNamesDict[gate.exchangeName]
        if gate.exchangeName.lower() == 'kucoin':
            extraParams = {'leverage': 5}
            futuresOrderData = gate.createAndTestFuturesOrder(symbolName, 'BUY', 'LIMIT', timeInForce='GTC',
                                                              price=20000, quantity=20, extraParams=extraParams)
        else:
            futuresOrderData = gate.createAndTestFuturesOrder(symbolName, 'BUY', 'LIMIT', timeInForce='GTC',
                                                              price=20000, quantity=0.002)
        result = gate.makeFuturesOrder(futuresOrderData)
        order = gate.getOrder(symbolName, orderId=result['orderId'], futures=True)

        # print('\nOrder data fetched from {} exchange: {}'.format(gate.exchangeName, order))

        assert order['clientOrderId'] == result[
            'clientOrderId'], 'Futures fetch client orderID is not equal to the actual client orderID from {} exchange.'.format(
            gate.exchangeName)

        order = gate.getOrder(symbolName, localOrderId=result['clientOrderId'], futures=True)
        assert order['orderId'] == result[
            'orderId'], 'Futures fetch orderID is not equal to the actual orderID from {} exchange.'.format(
            gate.exchangeName)


def testCancelingAllFuturesOpenOrders(getGatesAndSymbolNames):
    gates, symbolNamesDict = getGatesAndSymbolNames
    for gate in gates:
        symbolName = symbolNamesDict[gate.exchangeName]
        if gate.exchangeName.lower() == 'kucoin':
            extraParams = {'leverage': 5}
            futuresOrderData1 = gate.createAndTestFuturesOrder(symbolName, 'BUY', 'LIMIT', timeInForce='GTC',
                                                               price=20000, quantity=0.002, extraParams=extraParams)
            futuresOrderData2 = gate.createAndTestFuturesOrder(symbolName, 'BUY', 'LIMIT', timeInForce='GTC',
                                                               price=20000, quantity=0.002, extraParams=extraParams)
        else:
            futuresOrderData1 = gate.createAndTestFuturesOrder(symbolName, 'BUY', 'LIMIT', timeInForce='GTC',
                                                               price=20000, quantity=0.002)
            futuresOrderData2 = gate.createAndTestFuturesOrder(symbolName, 'BUY', 'LIMIT', timeInForce='GTC',
                                                               price=20000, quantity=0.002)
        gate.makeFuturesOrder(futuresOrderData1)
        gate.makeFuturesOrder(futuresOrderData2)

        gate.cancelAllSymbolOpenOrders(symbolName, futures=True)

        openOrders = gate.getOpenOrders(symbolName, futures=True)
        assert len(openOrders) == 0, 'Problem in canceling all Open Orders in {} exchange.'.format(gate.exchangeName)


def testCancelingOrder(getGatesAndSymbolNames):
    gates, symbolNamesDict = getGatesAndSymbolNames
    for gate in gates:
        symbolName = symbolNamesDict[gate.exchangeName]
        if gate.exchangeName.lower() == 'kucoin':
            extraParams = {'leverage': 5}
            futuresOrderData = gate.createAndTestFuturesOrder(symbolName, 'BUY', 'LIMIT', timeInForce='GTC',
                                                              price=20000, quantity=0.002, extraParams=extraParams,
                                                              newClientOrderId=str(int(time.time())))
        else:
            futuresOrderData = gate.createAndTestFuturesOrder(symbolName, 'BUY', 'LIMIT', timeInForce='GTC',
                                                              price=20000, quantity=0.002,
                                                              newClientOrderId=str(int(time.time())))

        result = gate.makeFuturesOrder(futuresOrderData)

        gate.cancelOrder(symbol=symbolName, orderId=result['orderId'], futures=True)

        result = gate.getOrder(symbol=symbolName, localOrderId=result['clientOrderId'], futures=True)
        assert result['status'].lower() in ['canceled', 'cancelled'], \
            'Problem in canceling specified open order by orderId from {} exchange.'.format(gate.exchangeName)

        if gate.exchangeName.lower() == 'kucoin':
            extraParams = {'leverage': 5}
            futuresOrderData = gate.createAndTestFuturesOrder(symbolName, 'BUY', 'LIMIT', timeInForce='GTC',
                                                              price=20000, quantity=0.002, extraParams=extraParams,
                                                              newClientOrderId=str(int(time.time())))
        else:
            futuresOrderData = gate.createAndTestFuturesOrder(symbolName, 'BUY', 'LIMIT', timeInForce='GTC',
                                                              price=20000, quantity=0.002,
                                                              newClientOrderId=str(int(time.time())))

        result = gate.makeFuturesOrder(futuresOrderData)

        gate.cancelOrder(symbol=symbolName, localOrderId=result['clientOrderId'], futures=True)

        result = gate.getOrder(symbol=symbolName, localOrderId=result['clientOrderId'], futures=True)
        assert result['status'].lower() in ['canceled', 'cancelled'], \
            'Problem in canceling specified open order by clientOrderId from {} exchange.'.format(gate.exchangeName)


def testFuturesTradeHistory(getGatesAndSymbolNames):
    gates, symbolNamesDict = getGatesAndSymbolNames
    for gate in gates:
        symbolName = symbolNamesDict[gate.exchangeName]
        tradeHistory = gate.symbolAccountTradeHistory(symbolName, futures=True)
        # print('\nTrade history from {} exchange: {}'.format(gate.exchangeName, tradeHistory))

        assert tradeHistory is not None, 'Problem in fetching trade history from {} exchange.'.format(gate.exchangeName)

        interface = ['symbol', 'id', 'orderId', 'orderListId', 'price', 'qty', 'quoteQty', 'commission',
                     'commissionAsset', 'time', 'isBuyer', 'isMaker', 'isBestMatch']

        errorMessage = 'Bad fetch trade history interface for {} exchange,'.format(gate.exchangeName)
        try:
            if gate.exchangeName != 'Binance':
                interface.append('exchangeSpecific')
                if sorted(list(tradeHistory[0].keys())) != sorted(interface):
                    assert False, errorMessage
            else:
                if sorted(list(tradeHistory[0].keys())) != sorted(interface):
                    assert False, errorMessage
        except Exception:
            assert False, errorMessage


def testFuturesSymbolList(getGatesAndSymbolNames):
    gates, symbolNamesDict = getGatesAndSymbolNames
    for gate in gates:
        if gate.exchangeName.lower() != 'binance':
            continue
        symbolList = gate.getSymbolList(futures=True)

        assert symbolList is not None



import time
from datetime import datetime

import pandas as pd

from Exchanges.BaseExchange import BaseExchange
from Utils import KuCoinHelpers, DataHelpers
from kucoin.client import User, Trade, Market
from kucoin_futures.client import FuturesUser, FuturesTrade, FuturesMarket


def is_symbol_status_valid(symbolName, symbolDatas, futures=False):
    if futures:
        for symbolData in symbolDatas:
            if symbolData.symbol == symbolName:
                if symbolData.status == "TRADING":
                    return True
                else:
                    return False
    else:
        for symbolData in symbolDatas:
            if symbolData["symbol"] == symbolName:
                if symbolData["enableTrading"]:
                    return True
                else:
                    return False
    return False


def checkSpotOrderDataValid(orderData: DataHelpers.OrderData):
    checkOrderSide(orderData, futures=False)
    checkOrderSymbol(orderData)
    checkOrderType(orderData, futures=False)

    if orderData.order_type == "market":
        checkOrderSize(orderData)
    else:
        if "limit" in orderData.order_type:
            checkOrderPrice(orderData)
            checkOrderSize(orderData)
        elif "stop" in orderData.order_type:
            checkStopOrderDatas(orderData)

        checkOrderTimeInForce(orderData)
        checkExtraParams(orderData)


def checkFuturesOrderDataValid(orderData):
    checkOrderSide(orderData, futures=True)
    checkOrderSymbol(orderData)
    checkOrderType(orderData, futures=True)

    if orderData.order_type == "market":
        checkOrderSize(orderData, futures=True)
    elif orderData.order_type == "limit":
        checkOrderPrice(orderData)
        checkOrderSize(orderData, futures=True)
        checkOrderTimeInForce(orderData, futures=True)
        checkExtraParams(orderData, futures=True)
    if orderData.stop_price is not None:
        checkStopOrderDatas(orderData, futures=True)


def checkExtraParams(orderData, futures=False):
    if orderData.extra_params is not None:
        if futures:
            checkPostOnlyOrder(orderData, futures)
            checkIceBergOrder(orderData)
            checkOrderLeverage(orderData)
        else:
            checkCancelAfterOrder(orderData)
            checkPostOnlyOrder(orderData, futures)


def checkCancelAfterOrder(orderData):
    if "cancelAfter" in orderData.extra_params.keys():
        if orderData.time_in_force != "GTT":
            raise ValueError(
                "'cancelAfter' field can only be used with 'GTT' as 'timeInForce' field."
            )


def checkOrderLeverage(orderData):
    if "leverage" not in orderData.extra_params.keys():
        if orderData.close_position is None:
            raise ValueError("Missing 'leverage' field.")
        if not orderData.close_position:
            raise ValueError("Missing 'leverage' field.")


def checkIceBergOrder(orderData):
    if "iceberg" in orderData.extra_params.keys():
        if "visibleSize" not in orderData.extra_params.keys():
            raise ValueError("Specify 'visibleSize' with 'iceberg' set as true")


def checkPostOnlyOrder(orderData, futures=False):
    if "postOnly" in orderData.extra_params.keys():
        if futures:
            if orderData.time_in_force in ["FOK"]:
                raise ValueError(
                    "'postOnly' field can not be used with 'IOC' as 'timeInForce' field."
                )
            if "hidden" in orderData.extra_params.keys():
                raise ValueError("Can't use 'hidden' with 'postOnly'")
            if "iceberg" in orderData.extra_params.keys():
                raise ValueError("Can't use 'iceberg' with 'postOnly'")
        else:
            if orderData.time_in_force in ["IOC", "FOK"]:
                raise ValueError(
                    "'postOnly' field can not be used with 'IOC' or 'FOK' as 'timeInForce' field."
                )


def checkOrderTimeInForce(orderData, futures=False):
    errorString = "Invalid value for 'timeInForce' specified"

    if futures:
        validValues = ["GTC", "IOC"]
    else:
        validValues = ["GTC", "GTT", "IOC", "FOK"]

    if orderData.time_in_force not in validValues:
        raise ValueError(errorString)


def checkOrderPrice(orderData):
    if orderData.price is None:
        raise ValueError("Missing 'price' field for limit order type.")


def checkStopOrderDatas(orderData, futures=False):
    if futures:
        if "stop" not in orderData.extra_params.keys():
            raise ValueError(
                "Specify 'stop' inside 'extraParams'. Either 'down' or 'up'."
            )
        if "stopPriceType" not in orderData.extra_params.keys():
            raise ValueError(
                "Specify 'stopPriceType' inside 'extraParams'. Either 'TP', 'IP' or 'MP'."
            )
    else:
        if orderData.extra_params is not None:
            if "stop" not in orderData.extra_params.keys():
                raise ValueError("Specify 'stop' in 'extraParams' for stop order.")
        else:
            raise ValueError("Specify 'stopPrice' in 'extraParams' for stop order.")

        if orderData.quantity is None:
            raise ValueError("Missing 'quantity' field for stop order type.")


def checkOrderSize(orderData, futures=False):
    errorString = "Provide either 'quantity' or 'quoteOrderQty'."
    if futures:
        if orderData.quantity is None and orderData.quote_quantity is None:
            if orderData.close_position is None:
                raise ValueError(errorString)
            if not orderData.close_position:
                raise ValueError(errorString)
    else:
        if orderData.quantity is None and orderData.quote_order_qty is None:
            raise ValueError(errorString)


def checkOrderType(orderData, futures=False):
    if futures:
        validTypes = ["limit", "market", "LIMIT", "MARKET", "Limit", "Market"]
    else:
        validTypes = [
            "limit",
            "market",
            "stop_market",
            "LIMIT",
            "MARKET",
            "STOP_MARKET",
            "Limit",
            "Market",
            "Stop_Market",
        ]

    if orderData.order_type is None or orderData.order_type not in validTypes:
        raise ValueError("Missing 'type' field.")
    orderData.order_type = orderData.order_type.lower()


def checkOrderSymbol(orderData):
    if orderData.symbol is None:
        raise ValueError("Missing 'symbol' field.")


def checkOrderSide(orderData, futures=False):
    errorString = "Missing or incorrect 'side' field."
    if futures:
        if orderData.side is None or orderData.side not in [
            "buy",
            "sell",
            "BUY",
            "SELL",
            "Buy",
            "Sell",
        ]:
            if orderData.close_position is None:
                raise ValueError(errorString)
            if not orderData.close_position:
                raise ValueError(errorString)
        if orderData.side is not None:
            orderData.side = orderData.side.lower()
    else:
        if orderData.side is None or orderData.side not in [
            "buy",
            "sell",
            "BUY",
            "SELL",
            "Buy",
            "Sell",
        ]:
            raise ValueError(errorString)
        orderData.side = orderData.side.lower()


class KuCoinExchange(BaseExchange):
    timeIntervals = [
        "1min",
        "3min",
        "5min",
        "15min",
        "30min",
        "1hour",
        "2hour",
        "4hour",
        "6hour",
        "8hour",
        "12hour",
        "1day",
        "1week",
    ]

    timeIntervalTranslate = {
        "1m": "1min",
        "3m": "3min",
        "5m": "5min",
        "15m": "15min",
        "30m": "30min",
        "1h": "1hour",
        "2h": "2hour",
        "4h": "4hour",
        "6h": "6hour",
        "8h": "8hour",
        "12h": "12hour",
        "1d": "1day",
        "1w": "1week",
    }

    noOrderIdsErrorString = (
        "Specify either 'orderId' or 'localOrderId' (only for active orders)"
    )

    def __init__(self, credentials, sandbox=False, unified_in_outs=True):
        self.spotApiKey = credentials["spot"]["key"]
        self.spotSecret = credentials["spot"]["secret"]
        self.spotPassphrase = credentials["spot"]["passphrase"]

        self.futuresApiKey = credentials["futures"]["key"]
        self.futuresSecret = credentials["futures"]["secret"]
        self.futuresPassphrase = credentials["futures"]["passphrase"]

        self.sandbox = sandbox
        self.unifiedInOuts = unified_in_outs

        self.unavailableErrorText = "This method is unavailable in KuCoin exchange"

        if sandbox:
            self.spotUser = User(
                key=self.spotApiKey,
                secret=self.spotSecret,
                passphrase=self.spotPassphrase,
                is_sandbox=True,
            )
            self.spotTrade = Trade(
                key=self.spotApiKey,
                secret=self.spotSecret,
                passphrase=self.spotPassphrase,
                is_sandbox=True,
            )
            self.spotMarket = Market(
                key=self.spotApiKey,
                secret=self.spotSecret,
                passphrase=self.spotPassphrase,
                is_sandbox=True,
            )

            self.futuresUser = FuturesUser(
                key=self.futuresApiKey,
                secret=self.futuresSecret,
                passphrase=self.futuresPassphrase,
                is_sandbox=True,
            )
            self.futuresTrade = FuturesTrade(
                key=self.futuresApiKey,
                secret=self.futuresSecret,
                passphrase=self.futuresPassphrase,
                is_sandbox=True,
            )
            self.futuresMarket = FuturesMarket(
                key=self.futuresApiKey,
                secret=self.futuresSecret,
                passphrase=self.futuresPassphrase,
                is_sandbox=True,
            )
        else:
            self.spotUser = User(
                key=self.spotApiKey,
                secret=self.spotSecret,
                passphrase=self.spotPassphrase,
            )
            self.spotTrade = Trade(
                key=self.spotApiKey,
                secret=self.spotSecret,
                passphrase=self.spotPassphrase,
            )
            self.spotMarket = Market(
                key=self.spotApiKey,
                secret=self.spotSecret,
                passphrase=self.spotPassphrase,
            )

            self.futuresUser = FuturesUser(
                key=self.futuresApiKey,
                secret=self.futuresSecret,
                passphrase=self.futuresPassphrase,
            )
            self.futuresTrade = FuturesTrade(
                key=self.futuresApiKey,
                secret=self.futuresSecret,
                passphrase=self.futuresPassphrase,
            )
            self.futuresMarket = FuturesMarket(
                key=self.futuresApiKey,
                secret=self.futuresSecret,
                passphrase=self.futuresPassphrase,
            )

    def get_balance(self, asset=None, futures=False):
        if futures:
            if asset is None:
                return KuCoinHelpers.unify_get_balance_futures_out(
                    [
                        self.futuresUser.get_account_overview(),
                        self.futuresUser.get_account_overview(currency="USDT"),
                    ]
                )
            else:
                return KuCoinHelpers.unify_get_balance_futures_out(
                    self.futuresUser.get_account_overview(currency=asset),
                    is_single=True,
                )

        else:
            if asset is None:
                return KuCoinHelpers.unify_get_balance_spot_out(
                    self.spotUser.get_account_list(currency=asset)
                )
            else:
                return KuCoinHelpers.unify_get_balance_spot_out(
                    self.spotUser.get_account_list(currency=asset), is_single=True
                )

    def symbol_account_trade_history(
        self, symbol, futures=False, from_id=None, limit=None
    ):
        if futures:
            return KuCoinHelpers.unify_trade_history(
                self.futuresTrade.get_fills_details(symbol=symbol)["items"],
                futures=True,
            )
        else:
            return KuCoinHelpers.unify_trade_history(
                self.spotTrade.get_fill_list(tradeType="TRADE")["items"]
            )

    def test_spot_order(self, order_data):
        checkSpotOrderDataValid(order_data)
        return order_data

    def make_spot_order(self, order_data):
        params = KuCoinHelpers.get_spot_order_as_dict(order_data)
        response = None

        if params["type"] == "market":
            response = self.spotTrade.create_market_order(**params)
        if params["type"] == "limit":
            response = self.spotTrade.create_limit_order(**params)
        if params["type"] == "stop_limit":
            params["type"] = "limit"
            response = self.spotTrade.create_limit_stop_order(**params)
        if params["type"] == "stop_market":
            params["type"] = "market"
            response = self.spotTrade.create_market_stop_order(**params)

        return self.get_order(
            params["symbol"], order_id=response["orderId"], futures=False
        )

    def create_and_test_spot_order(
        self,
        symbol,
        side,
        order_type,
        quantity=None,
        price=None,
        time_in_force=None,
        stop_price=None,
        iceberg_qty=None,
        new_order_resp_type=None,
        new_client_order_id=None,
        extra_params=None,
    ):
        currOrder = DataHelpers.set_spot_order_data(
            iceberg_qty,
            new_client_order_id,
            new_order_resp_type,
            order_type,
            price,
            quantity,
            side,
            stop_price,
            symbol,
            time_in_force,
            extra_params,
        )

        self.test_spot_order(currOrder)

        return currOrder

    def get_symbol_orders(
        self,
        symbol,
        futures=False,
        order_id=None,
        start_time=None,
        end_time=None,
        limit=None,
    ):
        if futures:
            args = {}
            if start_time is not None:
                args["startAt"] = start_time
            if end_time is not None:
                args["endAt"] = end_time
            args["symbol"] = symbol
            orderList = self.futuresTrade.get_order_list(**args)["items"]
            return KuCoinHelpers.unify_get_symbol_orders(orderList, futures=True)
        else:
            args = {}
            if start_time is not None:
                args["startAt"] = start_time
            if end_time is not None:
                args["endAt"] = end_time
            args["symbol"] = symbol
            orderList = self.spotTrade.get_order_list(**args)["items"]
            return KuCoinHelpers.unify_get_symbol_orders(orderList)

    def get_open_orders(self, symbol, futures=False):
        args = {"symbol": symbol, "status": "active"}
        if futures:
            lotSize = self.get_symbol_min_trade(symbol=symbol, futures=True)[
                "precisionStep"
            ]
            orderList = self.futuresTrade.get_order_list(**args)["items"]
            return KuCoinHelpers.unify_get_symbol_orders(
                orderList, futures=True, lot_size=lotSize
            )
        else:
            orderList = self.spotTrade.get_order_list(**args)["items"]
            return KuCoinHelpers.unify_get_symbol_orders(orderList)

    def cancel_all_symbol_open_orders(self, symbol, futures=False):
        if futures:
            result = self.futuresTrade.cancel_all_limit_order(symbol)[
                "cancelledOrderIds"
            ]
            result.append(
                self.futuresTrade.cancel_all_stop_order(symbol)["cancelledOrderIds"]
            )
            return result
        else:
            args = {"symbol": symbol}
            result = self.spotTrade.cancel_all_orders(**args)
            return result["cancelledOrderIds"]

    def cancel_order(self, symbol, order_id=None, local_order_id=None, futures=False):
        if futures:
            if order_id is not None:
                cancelledOrderId = self.futuresTrade.cancel_order(order_id)[
                    "cancelledOrderIds"
                ][0]
            elif local_order_id is not None:
                orderData = self.get_order(
                    symbol=symbol, local_order_id=local_order_id, futures=True
                )
                cancelledOrderId = self.futuresTrade.cancel_order(
                    orderId=orderData["orderId"]
                )["cancelledOrderIds"][0]
            else:
                raise ValueError(self.noOrderIdsErrorString)
            return self.get_order(symbol, order_id=cancelledOrderId, futures=True)
        else:
            if order_id is not None:
                cancelledOrderId = self.spotTrade.cancel_order(order_id)[
                    "cancelledOrderIds"
                ][0]
            elif local_order_id is not None:
                cancelledOrderId = self.spotTrade.cancel_client_order(local_order_id)[
                    "cancelledOrderId"
                ]
            else:
                raise ValueError(self.noOrderIdsErrorString)
            return self.get_order(symbol, order_id=cancelledOrderId, futures=False)

    def get_order(self, symbol, order_id=None, local_order_id=None, futures=False):
        if futures:
            if order_id is not None:
                orderData = self.futuresTrade.get_order_details(order_id)
            elif local_order_id is not None:
                orderData = self.futuresTrade.get_client_order_details(local_order_id)
            else:
                raise ValueError(self.noOrderIdsErrorString)

            lotSize = self.get_symbol_min_trade(symbol=symbol, futures=True)[
                "precisionStep"
            ]
            return KuCoinHelpers.unify_get_order(
                orderData, futures=True, lot_size=lotSize
            )
        else:
            if order_id is not None:
                orderData = self.spotTrade.get_order_details(order_id)
            elif local_order_id is not None:
                orderData = self.spotTrade.get_client_order_details(local_order_id)
            else:
                raise ValueError(self.noOrderIdsErrorString)

            return KuCoinHelpers.unify_get_order(orderData)

    def get_trading_fees(self, symbol=None, futures=False):
        if futures:
            if symbol is None:
                raise ValueError("Must specify futures contract symbol name.")
            contractInfo = self.futuresMarket.get_contract_detail(symbol=symbol)
            return {
                "symbol": contractInfo["symbol"],
                "takerCommission": contractInfo["takerFeeRate"],
                "makerCommission": contractInfo["makerFeeRate"],
            }
        else:
            if symbol is None:
                return self.spotUser.get_base_fee()
            else:
                feeData = self.spotUser.get_actual_fee(symbols=symbol)[0]

                return {
                    "symbol": feeData["symbol"],
                    "takerCommission": feeData["takerFeeRate"],
                    "makerCommission": feeData["makerFeeRate"],
                }

    def get_symbol_ticker_price(self, symbol, futures=False):
        if futures:
            return float(self.futuresMarket.get_ticker(symbol=symbol)["price"])
        else:
            return float(self.spotMarket.get_ticker(symbol=symbol)["price"])

    def get_symbol_klines(
        self,
        symbol,
        interval,
        start_time=None,
        end_time=None,
        limit=500,
        futures=False,
        blvtnav=False,
        convert_date_time=False,
        do_clean=False,
        to_clean_dataframe=False,
    ):
        if interval not in KuCoinExchange.timeIntervals:
            if interval in KuCoinExchange.timeIntervalTranslate.keys():
                timeInterval = KuCoinExchange.timeIntervalTranslate[interval]
            else:
                raise ValueError("Time interval is not valid.")
        else:
            timeInterval = interval

        if start_time is not None and not isinstance(start_time, int):
            start_time = int(
                datetime.timestamp(datetime.strptime(start_time, "%Y-%m-%d %H:%M:%S"))
            )
        if end_time is not None and not isinstance(end_time, int):
            end_time = int(
                datetime.timestamp(datetime.strptime(end_time, "%Y-%m-%d %H:%M:%S"))
            )

        if futures:
            data = self._getFuturesSymbolKlines(
                end_time, timeInterval, limit, start_time, symbol
            )
        else:
            if start_time is not None:
                start_time = int(str(start_time)[:-3])

            if end_time is not None:
                end_time = int(str(end_time)[:-3])

            data = self._getSpotSymbolKlines(
                end_time, timeInterval, limit, start_time, symbol
            )

        if convert_date_time or to_clean_dataframe:
            if futures:
                for datum in data:
                    datum.append(datum[-1])
                    datum[-1] = datetime.fromtimestamp((float(datum[0]) - 1) / 1000)
                    datum[0] = datetime.fromtimestamp(float(datum[0]) / 1000)
                    datum.append(None)
            else:
                for datum in data:
                    datum.append(datum[-1])
                    datum[-2] = datetime.fromtimestamp(float(datum[0]) - 1)
                    datum[0] = datetime.fromtimestamp(float(datum[0]))

        if do_clean or to_clean_dataframe:
            if to_clean_dataframe:
                if futures:
                    cleanDataFrame = pd.DataFrame(
                        data,
                        columns=[
                            "date",
                            "open",
                            "high",
                            "low",
                            "close",
                            "volume",
                            "closeDate",
                            "tradesNum",
                        ],
                    )
                else:
                    cleanDataFrame = pd.DataFrame(
                        data,
                        columns=[
                            "date",
                            "open",
                            "close",
                            "high",
                            "low",
                            "volume",
                            "closeDate",
                            "tradesNum",
                        ],
                    )
                cleanDataFrame.set_index("date", inplace=True)
                cleanDataFrame[cleanDataFrame.columns[:5]] = cleanDataFrame[
                    cleanDataFrame.columns[:5]
                ].apply(pd.to_numeric, errors="coerce")
                cleanDataFrame[cleanDataFrame.columns[-1]] = cleanDataFrame[
                    cleanDataFrame.columns[-1]
                ].apply(pd.to_numeric, errors="coerce")
                return cleanDataFrame
            return data
        else:
            return data

    def _getSpotSymbolKlines(self, endTime, timeInterval, limit, startTime, symbol):
        if limit is None:
            if startTime is None:
                if endTime is None:
                    data = self.spotMarket.get_kline(
                        symbol=symbol, kline_type=timeInterval
                    )
                else:
                    raise ValueError("Can't use endTime without limit.")
            else:
                if endTime is None:
                    data = self.spotMarket.get_kline(
                        symbol=symbol, kline_type=timeInterval, startAt=startTime
                    )
                else:
                    data = self.spotMarket.get_kline(
                        symbol=symbol,
                        kline_type=timeInterval,
                        startAt=startTime,
                        endAt=endTime,
                    )
        else:
            if startTime is None:
                if endTime is None:
                    startAt = int(time.time()) - limit * self._getTimeIntervalInSeconds(
                        timeInterval
                    )
                    data = self.spotMarket.get_kline(
                        symbol=symbol,
                        kline_type=timeInterval,
                        startAt=startAt,
                        endAt=int(time.time()),
                    )
                else:
                    startAt = endTime - limit * self._getTimeIntervalInSeconds(
                        timeInterval
                    )
                    data = self.spotMarket.get_kline(
                        symbol=symbol,
                        kline_type=timeInterval,
                        startAt=startAt,
                        endAt=endTime,
                    )
            else:
                if endTime is None:
                    endAt = startTime + limit * self._getTimeIntervalInSeconds(
                        timeInterval
                    )
                    data = self.spotMarket.get_kline(
                        symbol=symbol,
                        kline_type=timeInterval,
                        startAt=startTime,
                        endAt=endAt,
                    )
                else:
                    data = self.spotMarket.get_kline(
                        symbol=symbol,
                        kline_type=timeInterval,
                        startAt=startTime,
                        endAt=endTime,
                    )
        return data[::-1]

    def _getFuturesSymbolKlines(self, endTime, timeInterval, limit, startTime, symbol):
        granularity = int(self._getTimeIntervalInSeconds(timeInterval) / 60)
        if limit is None:
            if startTime is None:
                if endTime is None:
                    data = self.futuresMarket.get_kline_data(
                        symbol=symbol, granularity=granularity
                    )
                else:
                    endTime = endTime - endTime % (granularity * 60)
                    data = self.futuresMarket.get_kline_data(
                        symbol=symbol, granularity=granularity, end_t=endTime
                    )
            else:
                if endTime is None:
                    startTime = startTime - startTime % (granularity * 60)
                    data = self.futuresMarket.get_kline_data(
                        symbol=symbol, granularity=granularity, begin_t=startTime
                    )
                else:
                    endTime = endTime - endTime % (granularity * 60)
                    startTime = startTime - startTime % (granularity * 60)

                    data = self.futuresMarket.get_kline_data(
                        symbol=symbol,
                        granularity=granularity,
                        begin_t=startTime,
                        end_t=endTime,
                    )
        else:
            if startTime is None:
                if endTime is None:
                    endTime = int(time.time()) * 1000
                    endTime = endTime - endTime % (granularity * 60 * 1000)

                    startAt = endTime - limit * granularity * 60 * 1000
                    startAt = startAt - startAt % (granularity * 60 * 1000)

                    data = self.futuresMarket.get_kline_data(
                        symbol=symbol, granularity=granularity, begin_t=startAt
                    )
                else:
                    endTime = endTime - endTime % (granularity * 60 * 1000)

                    startTime = endTime - limit * granularity * 60 * 1000
                    startTime = startTime - startTime % (granularity * 60 * 1000)

                    data = self.futuresMarket.get_kline_data(
                        symbol=symbol,
                        granularity=granularity,
                        begin_t=startTime,
                        end_t=endTime,
                    )
            else:
                if endTime is None:
                    startTime = startTime - startTime % (granularity * 60 * 1000)

                    endTime = startTime + limit * granularity * 60 * 1000
                    endTime = endTime - endTime % (granularity * 60 * 1000)

                    data = self.futuresMarket.get_kline_data(
                        symbol=symbol,
                        granularity=granularity,
                        begin_t=startTime,
                        end_t=endTime,
                    )
                else:
                    startTime = startTime - startTime % (granularity * 60 * 1000)
                    endTime = endTime - endTime % (granularity * 60 * 1000)

                    data = self.futuresMarket.get_kline_data(
                        symbol=symbol,
                        granularity=granularity,
                        begin_t=startTime,
                        end_t=endTime,
                    )
        return data

    def _getTimeIntervalInSeconds(self, timeInterval):
        if timeInterval not in self.timeIntervals:
            raise ValueError("Time interval is not valid.")

        if timeInterval == "1min":
            return 60
        elif timeInterval == "3min":
            return 3 * 60
        elif timeInterval == "5min":
            return 5 * 60
        elif timeInterval == "15min":
            return 15 * 60
        elif timeInterval == "30min":
            return 40 * 60
        elif timeInterval == "1hour":
            return 3600
        elif timeInterval == "2hour":
            return 2 * 3600
        elif timeInterval == "4hour":
            return 4 * 3600
        elif timeInterval == "6hour":
            return 6 * 3600
        elif timeInterval == "8hour":
            return 8 * 3600
        elif timeInterval == "12hour":
            return 12 * 3600
        elif timeInterval == "1day":
            return 24 * 3600
        elif timeInterval == "1week":
            return 7 * 24 * 3600

    def get_exchange_time(self, futures=False):
        if futures:
            return self.futuresMarket.get_server_timestamp()
        else:
            return self.spotMarket.get_server_timestamp()

    def get_symbol_24h_ticker(self, symbol):
        return self.spotMarket.get_24h_stats(symbol)

    def test_futures_order(self, futures_order_data):
        checkFuturesOrderDataValid(futures_order_data)

    def make_futures_order(self, futures_order_data):
        if futures_order_data.quantity is None:
            if futures_order_data.quote_quantity is not None:
                lotSize = self.get_symbol_min_trade(
                    symbol=futures_order_data.symbol, futures=True
                )["precisionStep"]
                if futures_order_data.price is None:
                    currPrice = self.get_symbol_ticker_price(
                        futures_order_data.symbol, futures=True
                    )
                    futures_order_data.quantity = int(
                        round(futures_order_data.quote_quantity / currPrice / lotSize)
                    )
                else:
                    futures_order_data.quantity = int(
                        round(
                            futures_order_data.quote_quantity
                            / futures_order_data.price
                            / lotSize
                        )
                    )

        params = KuCoinHelpers.get_futures_order_as_dict(futures_order_data)

        symbol = params["symbol"]
        del params["symbol"]

        side = params["side"]
        del params["side"]

        if "leverage" in params.keys():
            leverage = params["leverage"]
            del params["leverage"]
        else:
            leverage = None
        if params["type"] == "market":
            result = self.futuresTrade.create_market_order(
                symbol, side, leverage, **params
            )
        elif params["type"] == "limit":
            size = params["size"]
            del params["size"]

            price = params["price"]
            del params["price"]

            result = self.futuresTrade.create_limit_order(
                symbol, side, leverage, size, price, **params
            )
        else:
            result = None
        return self.get_order(symbol, order_id=result["orderId"], futures=True)

    def create_and_test_futures_order(
        self,
        symbol,
        side,
        order_type,
        position_side=None,
        time_in_force=None,
        quantity=None,
        reduce_only=None,
        price=None,
        new_client_order_id=None,
        stop_price=None,
        close_position=None,
        activation_price=None,
        callback_rate=None,
        working_type=None,
        price_protect=None,
        new_order_resp_type=None,
        extra_params=None,
        quoteQuantity=None,
    ):
        currOrder = DataHelpers.set_futures_order_data(
            activation_price,
            callback_rate,
            close_position,
            extra_params,
            new_client_order_id,
            new_order_resp_type,
            order_type,
            position_side,
            price,
            price_protect,
            quantity,
            reduce_only,
            side,
            stop_price,
            symbol,
            time_in_force,
            working_type,
            quoteQuantity,
        )

        lotSize = self.get_symbol_min_trade(symbol=symbol, futures=True)[
            "precisionStep"
        ]
        if currOrder.quantity is not None:
            currOrder.quantity /= lotSize

        self.test_futures_order(currOrder)

        return currOrder

    def make_batch_futures_order(self, futures_order_datas):
        raise NotImplementedError(self.unavailableErrorText)

    def change_initial_leverage(self, symbol, leverage):
        raise NotImplementedError(self.unavailableErrorText)

    def change_margin_type(self, symbol, margin_type, params=None):
        if margin_type.upper() == "CROSSED":
            autoAdd = True
        elif margin_type.upper() == "ISOLATED":
            autoAdd = False
        else:
            raise ValueError(
                "Invalid value specified for 'marginType'. Must be either 'ISOLATED' or 'CROSSED'."
            )
        return self.futuresTrade.modify_auto_deposit_margin(symbol, autoAdd)["data"]

    def change_position_margin(self, symbol, amount):
        newPosition = self.futuresTrade.add_margin_manually(
            symbol=symbol, margin=amount, bizNo=str(time.time())
        )

        return True

    def get_position(self):
        return self.futuresTrade.get_all_position()

    def spot_best_bid_asks(self, symbol):
        tickerData = self.spotMarket.get_ticker(symbol)
        return KuCoinHelpers.unify_get_best_bid_asks(tickerData, symbol)

    def get_symbol_order_book(self, symbol, limit=None, futures=False):
        if futures:
            orderBook = self.futuresMarket.l2_order_book(symbol)
            return orderBook
        else:
            orderBook = self.spotMarket.get_aggregated_order(symbol)
            return orderBook

    def get_symbol_recent_trades(self, symbol, limit=None, futures=False):
        if futures:
            tradeHistory = self.futuresMarket.get_trade_history(symbol=symbol)
            return pd.DataFrame(
                KuCoinHelpers.unify_recent_trades(tradeHistory, futures=True)
            )
        else:
            tradeHistory = self.spotMarket.get_trade_histories(symbol=symbol)
            return pd.DataFrame(KuCoinHelpers.unify_recent_trades(tradeHistory))

    def get_position_info(self, symbol=None):
        if symbol is None:
            positionInfos = self.futuresTrade.get_all_position()
            return KuCoinHelpers.unify_get_position_infos(positionInfos)
        else:
            positionInfo = self.futuresTrade.get_position_details(symbol=symbol)
            return [KuCoinHelpers.unify_get_position_info(positionInfo)]

    def get_symbol_min_trade(self, symbol, futures=False):
        if futures:
            contractInfos = self.futuresMarket.get_contract_detail(symbol)
            return KuCoinHelpers.unify_min_trade(contractInfos, futures=True)
        else:
            symbolInfoList = self.spotMarket.get_symbol_list()

            symbolInfo = None
            for info in symbolInfoList:
                if info["symbol"] == symbol:
                    symbolInfo = info
            return KuCoinHelpers.unify_min_trade(symbolInfo)

    def make_sl_tp_limit_futures_order(
        self,
        symbol,
        order_side,
        quantity=None,
        quote_quantity=None,
        enter_price=None,
        take_profit=None,
        stop_loss=None,
        leverage=None,
        margin_type=None,
    ):
        symbolInfo = self.get_symbol_min_trade(symbol=symbol, futures=True)

        if quantity is None:
            if quote_quantity is None:
                raise ValueError("Specify either quantity or quoteQuantity")
            quantity = (
                int(quote_quantity / enter_price / symbolInfo["precisionStep"])
                * symbolInfo["precisionStep"]
            )

        if quantity < symbolInfo["minQuantity"]:
            raise ValueError("Quantity is lower than minimum quantity allowed.")

        mainOrder = self.create_and_test_futures_order(
            symbol,
            order_side.upper(),
            "LIMIT",
            quantity=quantity,
            price=enter_price,
            time_in_force="GTC",
            extra_params={"leverage": leverage},
        )

        tpSlSide = "sell" if order_side.upper() == "BUY" else "buy"

        slExtraParams = {
            "stop": "down" if order_side.upper() == "BUY" else "up",
            "stopPriceType": "TP",
        }
        stopLossOrder = self.create_and_test_futures_order(
            symbol=symbol,
            side=tpSlSide,
            order_type="MARKET",
            stop_price=stop_loss,
            close_position=True,
            time_in_force="GTC",
            extra_params=slExtraParams,
        )

        tpExtraParams = {
            "stop": "up" if order_side.upper() == "BUY" else "down",
            "stopPriceType": "TP",
        }
        takeProfitOrder = self.create_and_test_futures_order(
            symbol=symbol,
            side=tpSlSide,
            order_type="MARKET",
            stop_price=take_profit,
            close_position=True,
            time_in_force="GTC",
            extra_params=tpExtraParams,
        )

        mainOrderRes = self.make_futures_order(mainOrder)
        slOrderRes = self.make_futures_order(stopLossOrder)
        tpOrderRes = self.make_futures_order(takeProfitOrder)

        orderIds = {
            "mainOrder": mainOrderRes["orderId"],
            "stopLoss": slOrderRes["orderId"],
            "takeProfit": tpOrderRes["orderId"],
        }

        return orderIds

    def makeSlTpMarketFuturesOrder(
        self,
        symbol,
        orderSide,
        quantity=None,
        quoteQuantity=None,
        takeProfit=None,
        stopLoss=None,
        leverage=None,
        marginType=None,
    ):
        symbolInfo = self.get_symbol_min_trade(symbol=symbol, futures=True)
        marketPrice = self.get_symbol_ticker_price(symbol=symbol, futures=True)

        if leverage is None:
            raise ValueError("Must specify 'leverage' parameter for KuCoin orders.")

        if quantity is None:
            if quoteQuantity is None:
                raise ValueError("Specify either quantity or quoteQuantity")
            quantity = (
                int(quoteQuantity / marketPrice / symbolInfo["precisionStep"])
                * symbolInfo["precisionStep"]
            )

        if quantity < symbolInfo["minQuantity"]:
            raise ValueError("Quantity is lower than minimum quantity allowed.")

        mainOrder = self.create_and_test_futures_order(
            symbol,
            orderSide.upper(),
            "MARKET",
            quantity=quantity,
            extra_params={"leverage": leverage},
        )

        tpSlSide = "sell" if orderSide.upper() == "BUY" else "buy"

        slExtraParams = {
            "stop": "down" if orderSide.upper() == "BUY" else "up",
            "stopPriceType": "TP",
        }
        stopLossOrder = self.create_and_test_futures_order(
            symbol=symbol,
            side=tpSlSide,
            order_type="MARKET",
            stop_price=stopLoss,
            close_position=True,
            time_in_force="GTC",
            extra_params=slExtraParams,
        )

        tpExtraParams = {
            "stop": "up" if orderSide.upper() == "BUY" else "down",
            "stopPriceType": "TP",
        }
        takeProfitOrder = self.create_and_test_futures_order(
            symbol=symbol,
            side=tpSlSide,
            order_type="MARKET",
            stop_price=takeProfit,
            close_position=True,
            time_in_force="GTC",
            extra_params=tpExtraParams,
        )

        mainOrderRes = self.make_futures_order(mainOrder)
        slOrderRes = self.make_futures_order(stopLossOrder)
        tpOrderRes = self.make_futures_order(takeProfitOrder)

        orderIds = {
            "mainOrder": mainOrderRes["orderId"],
            "stopLoss": slOrderRes["orderId"],
            "takeProfit": tpOrderRes["orderId"],
        }

        return orderIds

    def getSymbol24hChanges(self, futures=False):
        changesList = []
        if futures:
            for ticker in self.futuresMarket.get_contracts_list():
                if ticker["status"] == "Open":
                    changesList.append(
                        (ticker["symbol"], float(ticker["priceChgPct"]) * 100)
                    )
        else:
            symbolInfos = self.spotMarket.get_symbol_list()
            for ticker in self.spotMarket.get_all_tickers()["ticker"]:
                if is_symbol_status_valid(ticker["symbol"], symbolInfos, futures=False):
                    changesList.append(
                        (ticker["symbol"], float(ticker["changeRate"]) * 100)
                    )

        return sorted(changesList, key=lambda x: x[1], reverse=True)

    def getSymbolList(self, futures=False):
        symbolNames = []
        if futures:
            for ticker in self.futuresMarket.get_contracts_list():
                symbolNames.append(ticker["symbol"])
        else:
            for ticker in self.spotMarket.get_all_tickers()["ticker"]:
                symbolNames.append(ticker["symbol"])

        return symbolNames

    def getLatestSymbolNames(self, numOfSymbols=None, futures=False):
        symbolDatas = []
        if futures:
            for symbolInfo in self.futuresMarket.get_contracts_list():
                symbolDatas.append(
                    (
                        symbolInfo["symbol"],
                        datetime.fromtimestamp(
                            float(symbolInfo["firstOpenDate"]) / 1000
                        ),
                    )
                )
                symbolDatas.sort(key=lambda x: x[1], reverse=True)
            if numOfSymbols is not None and numOfSymbols > len(symbolDatas):
                numOfSymbols = len(symbolDatas)
        else:
            raise NotImplementedError()
        return symbolDatas[:numOfSymbols]

    def getIncomeHistory(
        self, currency, incomeType=None, startTime=None, endTime=None, limit=None
    ):
        args = {
            "beginAt": startTime,
            "endAt": endTime,
            "type": incomeType,
            "maxCount": limit,
            "currency": currency,
        }
        args = {k: v for k, v in args.items() if v is not None}

        return KuCoinHelpers.unify_get_income(
            self.futuresUser.get_transaction_history(**args)["dataList"]
        )

    def get_long_short_ratios(
        self, symbol, period, limit=None, start_time=None, end_time=None
    ):
        raise NotImplementedError()

    def get_deposit_address(self, coin, network=None):
        raise NotImplementedError()

    def withdraw(self, coin, address, amount, extra_data):
        raise NotImplementedError()



class SymbolMiniTickerEvent:

    def __init__(self):
        self.eventType = ""
        self.eventTime = 0
        self.symbol = ""
        self.open = 0.0
        self.close = 0.0
        self.high = 0.0
        self.low = 0.0
        self.totalTradedBaseAssetVolume = 0.0
        self.totalTradedQuoteAssetVolume = 0.0

    @staticmethod
    def json_parse(json_wrapper):
        result = SymbolMiniTickerEvent()
        result.eventType = json_wrapper.get_string("e")
        result.eventTime = json_wrapper.get_int("E")
        result.symbol = json_wrapper.get_string("s")
        result.open = json_wrapper.get_float("o")
        result.close = json_wrapper.get_float("c")
        result.high = json_wrapper.get_float("h")
        result.low = json_wrapper.get_float("l")
        result.totalTradedBaseAssetVolume = json_wrapper.get_float("v")
        result.totalTradedQuoteAssetVolume = json_wrapper.get_float("q")
        return result


from kucoin_futures.base_request.base_request import KucoinFuturesBaseRestApi


class TradeData(KucoinFuturesBaseRestApi):

    def get_fund_history(self, symbol, startAt=None, endAt=None, reverse=True, offset=0, forward=True, maxCount=10):
        """
        https://docs.kumex.com/#get-funding-history

        :param symbol: interest symbol (Mandatory)
        :type: str
        :param startAt: start time(milisecond) (optional)
        :type: int
        :param endAt:  end time(milisecond) (optional)
        :type: int
        :param reverse: is reverse? (optional)
        :type: bool
        :param offset:  Start offset. The unique attribute of the last returned result of the last request. The data of (optional)
        the first page will be returned by default
        :type: int
        :param forward: his parameter functions to judge whether the lookup is forward or not (optional)
        :type: bool
        :param maxCount:Max record count (optional)
        :type : int
        :return:
        {
            "dataList": [
              {
                "id": 36275152660006,                //id
                "symbol": "XBTUSDM",                  //Symbol
                "timePoint": 1557918000000,          //Time point (milisecond)
                "fundingRate": 0.000013,             //Funding rate
                "markPrice": 8058.27,                //Mark price
                "positionQty": 10,                   //Position size
                "positionCost": -0.001241,           //Position value at settlement period
                "funding": -0.00000464               //Settled funding fees. A positive number means that the user received the funding fee, and vice versa.
                "settleCurrency": "XBT"             //settlement currency
              },
              .......
            ],
            "hasMore": true                         //Whether there are more pages
          }
        """

        params = {'symbol': symbol}

        if startAt:
            params['startAt'] = startAt
        if endAt:
            params['endAt'] = endAt
        if reverse:
            params['reverse'] = reverse
        if offset:
            params['offset'] = offset
        if forward:
            params['forward'] = forward
        if maxCount:
            params['maxCount'] = maxCount

        return self._request('GET', '/api/v1/funding-history', params=params)

    def get_position_details(self, symbol):
        """
        https://docs.kumex.com/#get-position-details

        :param symbol: interest symbol (Mandatory)
        :type: str
        :return:
        {
            "id": "5ce3cda60c19fc0d4e9ae7cd",                //Position ID
            "symbol": "XBTUSDM",                              //Symbol
            "autoDeposit": true,                             //Auto deposit margin or not
            "maintMarginReq": 0.005,                         //Maintenance margin requirement
            "riskLimit": 200,                                //Risk limit
            "realLeverage": 1.06,                            //Leverage of the order
            "crossMode": false,                              //Cross mode or not
            "delevPercentage": 0.1,                          //ADL ranking percentile
            "openingTimestamp": 1558433191000,               //Open time
            "currentTimestamp": 1558507727807,               //Current timestamp
            "currentQty": -20,                               //Current position
            "currentCost": 0.00266375,                       //Current position value
            "currentComm": 0.00000271,                       //Current commission
            "unrealisedCost": 0.00266375,                    //Unrealised value
            "realisedGrossCost": 0,                          //Accumulated realised gross profit value
            "realisedCost": 0.00000271,                      //Current realised position value
            "isOpen": true,                                  //Opened position or not
            "markPrice": 7933.01,                            //Mark price
            "markValue": 0.00252111,                         //Mark value
            "posCost": 0.00266375,                           //Position value
            "posCross": 1.2e-7,                              //Manually added margin
            "posInit": 0.00266375,                           //Leverage margin
            "posComm": 0.00000392,                           //Bankruptcy cost
            "posLoss": 0,                                    //Funding fees paid out
            "posMargin": 0.00266779,                         //Position margin
            "posMaint": 0.00001724,                          //Maintenance margin
            "maintMargin": 0.00252516,                       //Position margin
            "realisedGrossPnl": 0,                           //Accumulated realised gross profit value
            "realisedPnl": -0.00000253,                      //Realised profit and loss
            "unrealisedPnl": -0.00014264,                    //Unrealised profit and loss
            "unrealisedPnlPcnt": -0.0535,                    //Profit-loss ratio of the position
            "unrealisedRoePcnt": -0.0535,                    //Rate of return on investment
            "avgEntryPrice": 7508.22,                        //Average entry price
            "liquidationPrice": 1000000,                     //Liquidation price
            "bankruptPrice": 1000000                         //Bankruptcy price
            "settleCurrency": "XBT"                          //Currency used to clear and settle the trades
        }
        """
        params = {
            'symbol': symbol
        }
        return self._request('GET', '/api/v1/position', params=params)

    def get_all_position(self):
        """
        https://docs.kumex.com/#get-position-list
        :return:
        [{
          "id": "5ce3cda60c19fc0d4e9ae7cd",                //Position ID
          "symbol": "XBTUSDM",                              //Symbol
          "autoDeposit": true,                             //Auto deposit margin or not
          "maintMarginReq": 0.005,                         //Maintenance margin requirement
          "riskLimit": 200,                                //Risk limit
          "realLeverage": 1.06,                            //Leverage of the order
          "crossMode": false,                              //Cross mode or not
          "delevPercentage": 0.1,                          //ADL ranking percentile
          "openingTimestamp": 1558433191000,               //Open time
          "currentTimestamp": 1558507727807,               //Current timestamp
          "currentQty": -20,                               //Current position
          "currentCost": 0.00266375,                       //Current position value
          "currentComm": 0.00000271,                       //Current commission
          "unrealisedCost": 0.00266375,                    //Unrealised value
          "realisedGrossCost": 0,                          //Accumulated realised gross profit value
          "realisedCost": 0.00000271,                      //Current realised position value
          "isOpen": true,                                  //Opened position or not
          "markPrice": 7933.01,                            //Mark price
          "markValue": 0.00252111,                         //Mark value
          "posCost": 0.00266375,                           //Position value
          "posCross": 1.2e-7,                              //Manually added margin
          "posInit": 0.00266375,                           //Leverage margin
          "posComm": 0.00000392,                           //Bankruptcy cost
          "posLoss": 0,                                    //Funding fees paid out
          "posMargin": 0.00266779,                         //Position margin
          "posMaint": 0.00001724,                          //Maintenance margin
          "maintMargin": 0.00252516,                       //Position margin
          "realisedGrossPnl": 0,                           //Accumulated realised gross profit value
          "realisedPnl": -0.00000253,                      //Realised profit and loss
          "unrealisedPnl": -0.00014264,                    //Unrealised profit and loss
          "unrealisedPnlPcnt": -0.0535,                    //Profit-loss ratio of the position
          "unrealisedRoePcnt": -0.0535,                    //Rate of return on investment
          "avgEntryPrice": 7508.22,                        //Average entry price
          "liquidationPrice": 1000000,                     //Liquidation price
          "bankruptPrice": 1000000                         //Bankruptcy price
          "settleCurrency": "XBT"                         //Currency used to clear and settle the trades
        },
        ....]
        """
        return self._request('GET', '/api/v1/positions')

    def modify_auto_deposit_margin(self, symbol, status=True):
        """
        https://docs.kumex.com/#enable-disable-of-auto-deposit-margin

        :param  symbol: interest symbol (Mandatory)
        :param status: True is open, False is off  default True (Mandatory)
        :type: bool
        :return:  'True' or 'False'
        """
        params = {
            'symbol': symbol,
            'status': status
        }

        return self._request('POST', '/api/v1/position/margin/auto-deposit-status', params=params)

    def add_margin_manually(self, symbol, margin, bizNo):
        """
        https://docs.kumex.com/#add-margin-manually

        :param symbol:  (Mandatory)
        :type: str
        :param margin: Margin amount (min. margin amount≥0.00001667XBT）(Mandatory)
        :type: float
        :param bizNo: 	A unique ID generated by the user (Mandatory)
        :type: str
        :return:
        """
        params = {
            'symbol': symbol,
            'margin': margin,
            'bizNo': bizNo
        }
        return self._request('POST', '/api/v1/position/margin/deposit-margin', params=params)

    def get_contracts_risk_limit(self, symbol):
        """
        https://docs.kucoin.cloud/futures/#obtain-futures-risk-limit-level
        
        :param symbol:  (Mandatory)
        :type: str
        :return:
        """
        return self._request('GET', f'/api/v1/contracts/risk-limit/{symbol}')

    def change_position_risk_limit_level(self, symbol, level):
        """
        https://docs.kucoin.com/futures/#adjust-risk-limit-level
        
        :param symbol:  (Mandatory)
        :type: str
        :param level:  (Mandatory)
        :type: int
        :return:
        """
        params = {
            'symbol': symbol,
            'level': level,
        }
        return self._request('POST', '/api/v1/position/risk-limit-level/change', params=params)

    def get_fills_details(self, symbol='', orderId='', side='', type='', startAt=None, endAt=None, **kwargs):
        """
        https://docs.kumex.com/#get-fills
        :param symbol: [optional] Symbol of the contract
        :type: str
        :param orderId: List fills for a specific order only (If you specify orderId, other parameters can be ignored) [optional]
        :type: str
        :param side: [optional] buy or sell
        :type: str
        :param type: [optional] limit, market, limit_stop or market_stop
        :type: str
        :param startAt: [optional] Start time (milisecond)
        :type: int
        :param endAt: 	[optional] End time (milisecond)
        :type: str
        :param kwargs:  [optional]  currentPage , pageSize  and so on
        :return:
        {
      "currentPage":1,
      "pageSize":1,
      "totalNum":251915,
      "totalPage":251915,
      "items":[
          {
            "symbol": "XBTUSDM",  //Ticker symbol of the contract
            "tradeId": "5ce24c1f0c19fc3c58edc47c",  //Trade ID
            "orderId": "5ce24c16b210233c36ee321d",  // Order ID
            "side": "sell",  //Transaction side
            "liquidity": "taker",  //Liquidity- taker or maker
            "price": "8302",  //Filled price
            "size": 10,  //Filled amount
            "value": "0.001204529",  //Order value
            "feeRate": "0.0005",  //Floating fees
            "fixFee": "0.00000006",  //Fixed fees
            "feeCurrency": "XBT",  //Charging currency
            "stop": "",  //A mark to the stop order type
            "fee": "0.0000012022",  //Transaction fee
            "orderType": "limit",  //Order type
            "tradeType": "trade",  //Trade type (trade, liquidation or ADL)
            "createdAt": 1558334496000  //Time the order created
            "settleCurrency": "XBT", //settlement currency
            "tradeTime": 1558334496000000000 //trade time in nanosecond
          }
      ]
    }
        """
        params = {}
        if symbol:
            params['symbol'] = symbol
        if orderId:
            params['orderId'] = orderId
        if side:
            params['side'] = side
        if type:
            params['type'] = type
        if startAt:
            params['startAt'] = startAt
        if endAt:
            params['endAt'] = endAt
        if kwargs:
            params.update(kwargs)

        return self._request('GET', '/api/v1/fills', params=params)

    def get_recent_fills(self):
        """
        https://docs.kumex.com/#recent-fills

        :return:
        [{
         "symbol": "XBTUSDM",  //Ticker symbol of the contract
         "tradeId": "5ce24c1f0c19fc3c58edc47c",  //Trade ID
         "orderId": "5ce24c16b210233c36ee321d",  //Order ID
         "side": "sell",  //Transaction side
         "liquidity": "taker",  // Liquidity-taker or maker
         "price": "8302",  //Filled price
         "size": 10,  //Filled amount
         "value": "0.001204529",  //Order value
         "feeRate": "0.0005",  //Floating rate
         "fixFee": "0.00000006",  //Fixed fees
         "feeCurrency": "XBT",  //Charging currency
         "stop": "",  //A mark to the stop order type
         "fee": "0.0000012022",  //Transaction fee
         "orderType": "limit",  //Order type
         "tradeType": "trade",  //Trade type (tradek, liquidation or ADL )
         "createdAt": 1558334496000  //Time the order created
         "settleCurrency": "XBT", //settlement currency
         "tradeTime": 1558334496000000000 //trade time in nanosecond
        },....]
        """
        return self._request('GET', '/api/v1/recentFills')

    def get_open_order_details(self, symbol):
        """
        https://docs.kumex.com/#active-order-value-calculation

        :param symbol: interest symbol (Mandatory)
        :type: str
        :return:
        {'openOrderSellSize': 0, 'openOrderBuyCost': '0.170008023', 'openOrderBuySize': 1390, 'openOrderSellCost': '0'  "settleCurrency": "XBT" }
        """
        params = {
            'symbol': symbol,
        }

        return self._request('GET', '/api/v1/openOrderStatistics', params=params)

    def create_limit_order(self, symbol, side, lever, size, price, clientOid='', **kwargs):
        """
        Place Limit Order Functions

        https://docs.kumex.com/#place-an-order

        :param symbol: interest symbol (Mandatory)
        :type: str
        :param side: place direction buy or sell (Mandatory)
        :type: str
        :param lever: Leverage of the order (Mandatory)
        :type: str
        :param size: Order size. Must be a positive number (Mandatory)
        :type: str
        :param price: Limit price (Mandatory)
        :type: str
        :param clientOid: Unique order id created by users to identify their orders, e.g. UUID, Only allows numbers,
         characters, underline(_), and separator(-) (Mandatory)
        :type: str
        :param kwargs:  Fill in parameters with reference documents
        :return: {'orderId': '5d9ee461f24b80689797fd04'}
        """
        params = {
            'symbol': symbol,
            'size': size,
            'side': side,
            'price': price,
            'leverage': lever,
            'type': 'limit'
        }
        if not clientOid:
            clientOid = self.return_unique_id
        params['clientOid'] = clientOid
        if kwargs:
            params.update(kwargs)

        return self._request('POST', '/api/v1/orders', params=params)

    def create_market_order(self, symbol, side, lever, clientOid='', **kwargs):
        """
        Place Market Order Functions

        https://docs.kumex.com/#place-an-order

        :param symbol: interest symbol (Mandatory)
        :type: str
        :param side: place direction buy or sell (Mandatory)
        :type: str
        :param lever: Leverage of the order (Mandatory)
        :type: str
        :param clientOid: Unique order id created by users to identify their orders, e.g. UUID, Only allows numbers,
         characters, underline(_), and separator(-) (Mandatory)
        :type: str
        :param kwargs:  Fill in parameters with reference documents
        :return: {'orderId': '5d9ee461f24b80689797fd04'}
        """
        params = {
            'symbol': symbol,
            'side': side,
            'leverage': lever,
            'type': 'market'

        }
        if not clientOid:
            clientOid = self.return_unique_id
        params['clientOid'] = clientOid
        if kwargs:
            params.update(kwargs)

        return self._request('POST', '/api/v1/orders', params=params)

    def cancel_order(self, orderId):
        """
        https://docs.kumex.com/#cancel-an-order

        :param orderId: str  (Mandatory)
        :return:{'cancelledOrderIds': ['5d9ee77825aa3809494eac87']}
        """
        return self._request('DELETE', f'/api/v1/orders/{orderId}')

    def cancel_all_limit_order(self, symbol):
        """
        https://docs.kumex.com/#limit-order-mass-cancelation

        :param symbol: str  (Mandatory)
        :return:{'cancelledOrderIds': ['5d9d684ef24b806897632962', '5d9dd98725aa380949298eb0', ...]}
        """
        params = {
            'symbol': symbol
        }
        return self._request('DELETE', '/api/v1/orders', params=params)

    def cancel_all_stop_order(self, symbol):
        """
        https://docs.kumex.com/#stop-order-mass-cancelation

        :param symbol: str  (Mandatory)
        :return:{'cancelledOrderIds': ['5d9d684ef24b806897632962', '5d9dd98725aa380949298eb0', ...]}
        """
        params = {
            'symbol': symbol
        }
        return self._request('DELETE', '/api/v1/stopOrders', params=params)

    def get_order_list(self, **kwargs):
        """
        https://docs.kumex.com/#get-order-list

         return List your current orders.

        :param kwargs: [optional] symbol, status, side, type, startAt, endAt, currentPage , pageSize  and so on
        :return:
        {
          "currentPage": 1,
          "pageSize": 100,
          "totalNum": 1000,
          "totalPage": 10,
          "items": [
            {
              "id": "5cdfc138b21023a909e5ad55", //Order ID
              "symbol": "XBTUSDM",  //Ticker symbol of the contract
              "type": "limit",   //Order type, market order or limit order
              "side": "buy",  //Transaction side
              "price": "3600",  //Order price
              "size": 20000,  //Order quantity
              "value": "56.1167227833",  //Order value
              "dealValue": "0",  //Value of the executed orders
              "dealSize": 0,  //Executed order quantity
              "stp": "",  //Self trade prevention types
              "stop": "",  //Stop order type (stop limit or stop market)
              "stopPriceType": "",  //Trigger price type of stop orders
              "stopTriggered": false,  //Mark to show whether the stop order is triggered
              "stopPrice": null,  //Trigger price of stop orders
              "timeInForce": "GTC",  //Time in force policy type
              "postOnly": false,  // Mark of post only
              "hidden": false,  //Mark of the hidden order
              "iceberg": false,  //Mark of the iceberg order
              "visibleSize": null,  //Visible size of the iceberg order
              "leverage": "20",  //Leverage of the order
              "forceHold": false,  //A mark to forcely hold the funds for an order
              "closeOrder": false, //A mark to close the position
              "reduceOnly": false,  //A mark to reduce the position size only
              "clientOid": "5ce24c16b210233c36ee321d",  //Unique order id created by users to identify their orders
              "remark": null,  //Remark of the order
              "isActive": true,  //Mark of the active orders
              "cancelExist": false,  //Mark of the canceled orders
              "createdAt": 1558167872000  //Time the order created
              "settleCurrency": "XBT", //settlement currency
              status": "open", //order status: “open” or “done”
             "updatedAt": 1558167872000, //last update time
              "orderTime": 1558167872000000000 //order create time in nanosecond
            }
          ]
        }
        """
        params = {}
        if kwargs:
            params.update(kwargs)

        return self._request('GET', '/api/v1/orders', params=params)

    def get_open_stop_order(self, **kwargs):
        """
        https://docs.kumex.com/#get-untriggered-stop-order-list

        :param kwargs:c
        :return:
        {
      "currentPage": 1,
      "pageSize": 100,
      "totalNum": 1000,
      "totalPage": 10,
      "items": [
        {
            "id": "5cdfc138b21023a909e5ad55", //Order ID
            "symbol": "XBTUSDM",  //Ticker symbol of the contract
            "type": "limit",   //Order type, market order or limit order
            "side": "buy",  //Transaction side
            "price": "3600",  //Order price
            "size": 20000,  //Order quantity
            "value": "56.1167227833",  //Order value
            "dealValue": "0",  //Value of the executed orders
            "dealSize": 0,  //Executed order quantity
            "stp": "",  //Self trade prevention types
            "stop": "",  //Stop order type (stop limit or stop market)
            "stopPriceType": "",  //Trigger price type of stop orders
            "stopTriggered": false,  //Mark to show whether the stop order is triggered
            "stopPrice": null,  //Trigger price of stop orders
            "timeInForce": "GTC",  //Time in force policy type
            "postOnly": false,  //Mark of post only
            "hidden": false,  //Mark of the hidden order
            "iceberg": false,  //Mark of the iceberg order
            "visibleSize": null,  //Visible size of the iceberg order
            "leverage": "20",  //Leverage of the order
            "forceHold": false,  //A mark to forcely hold the funds for an order
            "closeOrder": false, //A mark to close the position
            "reduceOnly": false,  //A mark to reduce the position size only
            "clientOid": "5ce24c16b210233c36ee321d",  //Unique order id created by users to identify their orders
            "remark": null,  //Remark of the order
            "isActive": true,  //Mark of the active orders
            "cancelExist": false,  //Mark of the canceled orders
            "createdAt": 1558167872000  //Time the order created
            "settleCurrency": "XBT", //settlement currency
            "status": "open", //order status: “open” or “done”
            "updatedAt": 1558167872000 //last update time
        }
      ]
    }
        """
        params = {}
        if kwargs:
            params.update(kwargs)

        return self._request('GET', '/api/v1/stopOrders', params=params)

    def get_24h_done_order(self, **kwargs):
        """
        https://docs.kumex.com/#get-list-of-orders-completed-in-24h
        :param kwargs: [optional] currentPage , pageSize  and so on
        :return:
         {
      "currentPage": 1,
      "pageSize": 1,
      "totalNum": 153408,
      "totalPage": 15340,
      "items": [
          {
            "id": "5cdfc138b21023a909e5ad55", //Order ID
            "symbol": "XBTUSDM",  //Ticker symbol of the contract
            "type": "limit",   //Order type, market order or limit order
            "side": "buy",  //Transaction side
            "price": "3600",  //Order price
            "size": 20000,  //Order quantity
            "value": "56.1167227833",  //Order value
            "dealValue": "56.1167227833",  //Value of the executed orders
            "dealSize": 20000,  //Executed order quantity
            "stp": "",  //Self trade prevention types
            "stop": "",  //Stop order type (stop limit or stop market)
            "stopPriceType": "",  //Trigger price type of stop orders
            "stopTriggered": true,  //Mark to show whether the stop order is triggered
            "stopPrice": null,  //Trigger price of stop orders
            "timeInForce": "GTC",  //Time in force policy type
            "postOnly": false,  //Mark of post only
            "hidden": false,  //Mark of the hidden order
            "iceberg": false,  //Mark of the iceberg order
            "visibleSize": null,  // Visible size of the iceberg order
            "leverage": "20",  //Leverage of the order
            "forceHold": false,  //A mark to forcely hold the funds for an order
            "closeOrder": false, //A mark to close the position
            "reduceOnly": false,  //A mark to reduce the position size only
            "clientOid": "5ce24c16b210233c36ee321d",  //Unique order id created by users to identify their orders
            "remark": null,  //Remark of the order
            "isActive": false,  //Mark of the active orders
            "cancelExist": false,  //Mark of the canceled orders
            "createdAt": 1558167872000  //Time the order created
            "settleCurrency": "XBT", //settlement currency
            "status": "done", //order status: “open” or “done”
            "updatedAt": 1558167872000, //last update time
            "orderTime": 1558167872000000000 //order create time in nanosecond
          }
      ]
    }
        """
        params = {}
        if kwargs:
            params.update(kwargs)

        return self._request('GET', '/api/v1/recentDoneOrders', params=params)

    def get_order_details(self, orderId):
        """
        https://docs.kumex.com/#get-details-of-a-single-order
        Get a single order by order id (including a stop order).
        :param orderId: str
        :return:
        {
          "id": "5cdfc138b21023a909e5ad55", //Order ID
          "symbol": "XBTUSDM",  //Ticker symbol of the contract
          "type": "limit",   //Order type, market order or limit order
          "side": "buy",  //Transaction side
          "price": "3600",  //Order price
          "size": 20000,  //Order quantity
          "value": "56.1167227833",  //Order value
          "dealValue": "56.1167227833",  //Value of the executed orders
          "dealSize": 20000,  //Executed order quantity
          "stp": "",  //Self trade prevention types
          "stop": "",  //Stop order type (stop limit or stop market)
          "stopPriceType": "",  //Trigger price type of stop orders
          "stopTriggered": true,  //Mark to show whether the stop order is triggered
          "stopPrice": null,  //Trigger price of stop orders
          "timeInForce": "GTC",  //Time in force policy types
          "postOnly": false,  //Mark of post only
          "hidden": false,  //Mark of the hidden order
          "iceberg": false,  //Mark of the iceberg order
          "visibleSize": null,  //Visible size of the iceberg order
          "leverage": "20",  //Leverage of the order
          "forceHold": false,  //A mark to forcely hold the funds for an order
          "closeOrder": false, //A mark to close the position
          "reduceOnly": false,  //A mark to reduce the position size only
          "clientOid": "5ce24c16b210233c36ee321d",  //Unique order id created by users to identify their orders
          "remark": null,  //Remarks of the order
          "isActive": false,  //Mark of the active orders
          "cancelExist": false,  //Mark of the canceled orders
          "createdAt": 1558167872000  //Time the order created
          "settleCurrency": "XBT", //settlement currency
          "status": "done", //order status: “open” or “done”
          "updatedAt": 1558167872000, //last update time
          "orderTime": 1558167872000000000 //order create time in nanosecond
        }
        """
        return self._request('GET', f'/api/v1/orders/{orderId}')

    def get_client_order_details(self, clientOid):
        """
        https://docs.kumex.com/#get-details-of-a-single-order
        Get a single order by client order id (including a stop order).
        :param clientOid: str
        :return:
        {
          "id": "5cdfc138b21023a909e5ad55", //Order ID
          "symbol": "XBTUSDM",  //Ticker symbol of the contract
          "type": "limit",   //Order type, market order or limit order
          "side": "buy",  //Transaction side
          "price": "3600",  //Order price
          "size": 20000,  //Order quantity
          "value": "56.1167227833",  //Order value
          "dealValue": "56.1167227833",  //Value of the executed orders
          "dealSize": 20000,  //Executed order quantity
          "stp": "",  //Self trade prevention types
          "stop": "",  //Stop order type (stop limit or stop market)
          "stopPriceType": "",  //Trigger price type of stop orders
          "stopTriggered": true,  //Mark to show whether the stop order is triggered
          "stopPrice": null,  //Trigger price of stop orders
          "timeInForce": "GTC",  //Time in force policy types
          "postOnly": false,  //Mark of post only
          "hidden": false,  //Mark of the hidden order
          "iceberg": false,  //Mark of the iceberg order
          "visibleSize": null,  //Visible size of the iceberg order
          "leverage": "20",  //Leverage of the order
          "forceHold": false,  //A mark to forcely hold the funds for an order
          "closeOrder": false, //A mark to close the position
          "reduceOnly": false,  //A mark to reduce the position size only
          "clientOid": "5ce24c16b210233c36ee321d",  //Unique order id created by users to identify their orders
          "remark": null,  //Remarks of the order
          "isActive": false,  //Mark of the active orders
          "cancelExist": false,  //Mark of the canceled orders
          "createdAt": 1558167872000  //Time the order created
          "settleCurrency": "XBT", //settlement currency
          "status": "done", //order status: “open” or “done”
          "updatedAt": 1558167872000, //last update time
          "orderTime": 1558167872000000000 //order create time in nanosecond
        }
        """
        return self._request('GET', f'/api/v1/orders/byClientOid?clientOid={clientOid}')



class OrderUpdate:
    def __init__(self):
        self.eventType = ""
        self.eventTime = 0
        self.transactionTime = 0
        self.symbol = ""
        self.clientOrderId = ""
        self.side = None
        self.type = None
        self.timeInForce = None
        self.origQty = 0.0
        self.price = 0.0
        self.avgPrice = 0.0
        self.stopPrice = 0.0
        self.executionType = ""
        self.orderStatus = ""
        self.orderId = None
        self.lastFilledQty = 0.0
        self.cumulativeFilledQty = 0.0
        self.lastFilledPrice = 0.0
        self.commissionAsset = None
        self.commissionAmount = 0
        self.orderTradeTime = 0
        self.tradeID = None
        self.bidsNotional = 0.0
        self.asksNotional = 0.0
        self.isMarkerSide = None
        self.isReduceOnly = None
        self.workingType = 0.0
        self.isClosePosition = None
        self.activationPrice = 0.0
        self.callbackRate = 0.0
        self.positionSide = None


    @staticmethod
    def json_parse(json_data):
        result = OrderUpdate()
        result.eventType = json_data.get_string("e")
        result.eventTime = json_data.get_int("E")
        result.transactionTime = json_data.get_int("T")

        data_group = json_data.get_object("o")
        result.symbol = data_group.get_string("s")
        result.clientOrderId = data_group.get_string("c")
        result.side = data_group.get_string("S")
        result.type = data_group.get_string("o")
        result.timeInForce = data_group.get_string("f")
        result.origQty = data_group.get_float("q")
        result.price = data_group.get_float("p")
        result.avgPrice = data_group.get_float("ap")
        result.stopPrice = data_group.get_float("sp")
        result.executionType = data_group.get_string("x")
        result.orderStatus = data_group.get_string("X")
        result.orderId = data_group.get_int("i")
        result.lastFilledQty = data_group.get_float("l")
        result.cumulativeFilledQty = data_group.get_float("z")
        result.lastFilledPrice = data_group.get_float("L")
        result.commissionAsset = data_group.get_string_or_default("N", None)
        result.commissionAmount = data_group.get_float_or_default("n", None)
        result.orderTradeTime = data_group.get_int("T")
        result.tradeID = data_group.get_int("t")
        result.bidsNotional = data_group.get_float("b")
        result.asksNotional = data_group.get_float("a")
        result.isMarkerSide = data_group.get_boolean("m")
        result.isReduceOnly = data_group.get_boolean("R")
        result.workingType = data_group.get_string("wt")
        result.isClosePosition = data_group.get_boolean("cp")
        result.activationPrice = data_group.get_float_or_default("AP", None)
        result.callbackRate = data_group.get_float_or_default("cr", None)
        result.positionSide = data_group.get_string("ps")

        return result



import json
from datetime import datetime

import pandas as pd

from Utils import DataHelpers


def get_spot_order_as_dict(order: DataHelpers.OrderData):
    if order.timestamp is None:
        raise ValueError("Timestamp must be set")

    params = {
        "symbol": order.symbol,
        "side": order.side,
        "type": order.order_type,
        "timestamp": order.timestamp,
    }

    if order.time_in_force is not None:
        params["timeInForce"] = order.time_in_force

    if order.quantity is not None:
        params["quantity"] = order.quantity

    if order.quote_order_qty is not None:
        params["quoteOrderQty"] = order.quote_order_qty

    if order.price is not None:
        params["price"] = order.price

    if order.new_order_resp_type is not None:
        params["newOrderRespType"] = order.new_order_resp_type

    if order.stop_price is not None:
        params["stopPrice"] = order.stop_price

    if order.iceberg_qty is not None:
        params["icebergQty"] = order.iceberg_qty

    if order.new_client_order_id is not None:
        params["newClientOrderId"] = order.new_client_order_id

    if order.recv_window is not None:
        params["recvWindow"] = order.recv_window

    return params


def get_futures_order_as_dict(order: DataHelpers.FuturesOrderData, all_str=False):
    params = {"symbol": order.symbol, "side": order.side, "ordertype": order.orderType}

    if order.position_side is not None:
        params["positionSide"] = order.position_side

    if order.time_in_force is not None:
        params["timeInForce"] = order.time_in_force

    if order.quantity is not None:
        params["quantity"] = order.quantity

    if order.reduce_only is not None:
        params["reduceOnly"] = order.reduce_only

    if order.price is not None:
        params["price"] = order.price

    if order.new_client_order_id is not None:
        params["newClientOrderId"] = order.new_client_order_id

    if order.stop_price is not None:
        params["stopPrice"] = order.stop_price

    if order.close_position is not None:
        params["closePosition"] = order.close_position

    if order.activation_price is not None:
        params["activationPrice"] = order.activation_price

    if order.callback_rate is not None:
        params["callbackRate"] = order.callback_rate

    if order.working_type is not None:
        params["workingType"] = order.working_type

    if order.price_protect is not None:
        params["priceProtect"] = order.price_protect

    if order.new_order_resp_type is not None:
        params["newOrderRespType"] = order.new_order_resp_type

    if all_str:
        for key, value in params.items():
            params[key] = str(value)

    return params


def get_klines_desired_only_cols(data, desired_indexes):
    return [[datum[index] for index in desired_indexes] for datum in data]


def klines_convert_to_pandas(out_array):
    df = pd.DataFrame(
        out_array,
        columns=[
            "date",
            "open",
            "high",
            "low",
            "close",
            "volume",
            "closeDate",
            "tradesNum",
        ],
    )
    df.set_index("date", inplace=True)
    return df


def klines_convert_date(data, time_col_idxs):
    for datum in data:
        for idx in time_col_idxs:
            datum[idx] = datetime.fromtimestamp(float(datum[idx]) / 1000)


def extract_symbol_info_from_filters(symbol_filters, ticker_price):
    params = {}
    for symbolFilter in symbol_filters:
        if symbolFilter["filterType"] == "LOT_SIZE":
            params["minQuantity"] = float(symbolFilter["minQty"])
            params["precisionStep"] = float(symbolFilter["stepSize"])
            params["minQuoteQuantity"] = ticker_price * params["minQuantity"]

        if symbolFilter["filterType"] == "PRICE_FILTER":
            params["stepPrice"] = float(symbolFilter["tickSize"])
    return params


def make_batch_order_data(futures_order_datas):
    batch_orders = []
    for order in futures_order_datas:
        order_as_dict = get_futures_order_as_dict(order, all_str=True)
        order_as_dict["type"] = order_as_dict.pop("ordertype")

        order_json = json.dumps(order_as_dict)

        batch_orders.append(order_json)
    return batch_orders



import pandas as pd


def unify_get_balance_spot_out(data, is_single=False):
    all_assets = []

    for asset in data:
        asset_index = get_asset_index_in_list(asset["currency"], all_assets)
        if asset_index == -1:
            asset_info = new_empty_asset(asset["currency"])
        else:
            asset_info = all_assets[asset_index]

        asset_info["free"] += float(asset["available"])
        asset_info["locked"] += float(asset["holds"])
        asset_info["exchangeSpecific"].append(asset)

        if asset_index == -1:
            all_assets.append(asset_info)

    return all_assets[0] if is_single else all_assets


def new_empty_asset(asset_name):
    return {"asset": asset_name, "free": 0.0, "locked": 0.0, "exchangeSpecific": []}


def get_asset_index_in_list(asset_name, all_assets):
    return next(
        (i for i in range(len(all_assets)) if asset_name == all_assets[i]["asset"]),
        -1,
    )


def unify_trade_history(trade_history, futures=False):
    unified_trade_history = []

    for trade in trade_history:
        is_buyer = trade["liquidity"] == "taker"
        is_maker = trade["liquidity"] == "maker"
        if futures:
            unified_trade_history.append(
                {
                    "symbol": trade["symbol"],
                    "id": trade["tradeId"],
                    "orderId": trade["orderId"],
                    "orderListId": -1,
                    "price": trade["price"],
                    "qty": trade["value"],
                    "quoteQty": trade["size"],
                    "commission": trade["fee"],
                    "commissionAsset": trade["feeCurrency"],
                    "time": trade["tradeTime"],
                    "isBuyer": is_buyer,
                    "isMaker": is_maker,
                    "isBestMatch": None,
                    "exchangeSpecific": trade,
                }
            )
        else:
            unified_trade_history.append(
                {
                    "symbol": trade["symbol"],
                    "id": trade["tradeId"],
                    "orderId": trade["orderId"],
                    "orderListId": -1,
                    "price": trade["price"],
                    "qty": trade["size"],
                    "quoteQty": trade["funds"],
                    "commission": trade["fee"],
                    "commissionAsset": trade["feeCurrency"],
                    "time": trade["createdAt"],
                    "isBuyer": is_buyer,
                    "isMaker": is_maker,
                    "isBestMatch": None,
                    "exchangeSpecific": trade,
                }
            )

    return pd.DataFrame(unified_trade_history)


def unify_recent_trades(trade_history, futures=False):
    unified_trade_history = []

    for trade in trade_history:
        if futures:
            unified_trade_history.append(
                {
                    "id": int(trade["sequence"]),
                    "price": float(trade["price"]),
                    "qty": float(trade["size"]),
                    "quoteQty": float(trade["price"]) * float(trade["size"]),
                    "time": int(trade["ts"] / 1000),
                    "isBuyerMaker": None,
                    "exchangeSpecific": trade,
                }
            )
        else:
            unified_trade_history.append(
                {
                    "id": int(trade["sequence"]),
                    "price": float(trade["price"]),
                    "qty": float(trade["size"]),
                    "quoteQty": float(trade["price"]) * float(trade["size"]),
                    "time": int(trade["time"] / 1000),
                    "isBuyerMaker": None,
                    "isBestMatch": None,
                    "exchangeSpecific": trade,
                }
            )

    return unified_trade_history


def get_spot_order_as_dict(order_data):
    params = {
        "side": order_data.side,
        "symbol": order_data.symbol,
        "type": order_data.order_type,
    }

    if order_data.new_client_order_id is not None:
        params["clientOid"] = order_data.new_client_order_id

    if order_data.price is not None:
        params["price"] = order_data.price

    if order_data.quantity is not None:
        params["size"] = order_data.quantity

    if order_data.time_in_force is not None:
        params["timeInForce"] = order_data.time_in_force

    if order_data.quote_order_qty is not None and "size" not in params.keys():
        params["funds"] = order_data.quote_order_qty

    if order_data.extra_params is not None:
        set_extra_params(order_data, params)

    return params


def set_extra_params(order_data, params):
    if "cancelAfter" in order_data.extra_params.keys():
        params["cancelAfter"] = order_data.extra_params["cancelAfter"]
    if "postOnly" in order_data.extra_params.keys():
        params["postOnly"] = order_data.extra_params["postOnly"]
    if "hidden" in order_data.extra_params.keys():
        params["hidden"] = order_data.extra_params["hidden"]
    if "iceberg" in order_data.extra_params.keys():
        params["iceberg"] = order_data.extra_params["iceberg"]
    if "visibleSize" in order_data.extra_params.keys():
        params["visibleSize"] = order_data.extra_params["visibleSize"]
    if "stopPrice" in order_data.extra_params.keys():
        params["stopPrice"] = order_data.extra_params["stopPrice"]


def unify_get_order(order_data, futures=False, lot_size=None):
    if futures:
        if (
            order_data["value"] is not None
            and order_data["price"] is None
            and float(order_data["size"]) > 0
        ):
            order_data["price"] = float(order_data["value"]) / (
                float(order_data["size"]) * lot_size
            )

        if order_data["cancelExist"]:
            order_status = "CANCELLED"
        elif order_data["isActive"]:
            order_status = "NEW"
        else:
            order_status = "FILLED"

        return {
            "symbol": order_data["symbol"],
            "orderId": order_data["id"],
            "clientOrderId": order_data["clientOid"],
            "transactTime": order_data["createdAt"],
            "price": order_data["price"],
            "origQty": float(order_data["size"]) * lot_size,
            "executedQty": float(order_data["filledSize"]) * lot_size,
            "cummulativeQuoteQty": 0
            if order_data["price"] is None
            else float(order_data["filledSize"])
            * float(order_data["price"])
            * lot_size,
            "status": order_status,
            "timeInForce": order_data["timeInForce"],
            "type": order_data["type"],
            "side": order_data["side"],
            "extraData": {
                "reduceOnly": order_data["reduceOnly"],
                "stopPrice": 0.0
                if order_data["stopPrice"] is None
                else float(order_data["stopPrice"]),
                "workingType": "CONTRACT_PRICE",
                "avgPrice": order_data["price"],
                "origType": order_data["type"],
                "positionSide": "BOTH",
                "activatePrice": None,
                "priceRate": None,
                "closePosition": order_data["closeOrder"],
            },
            "exchangeSpecific": order_data,
        }
    else:
        if order_data["cancelExist"]:
            order_status = "CANCELLED"
        elif order_data["isActive"]:
            order_status = "NEW"
        else:
            order_status = "FILLED"

        return {
            "symbol": order_data["symbol"],
            "orderId": order_data["id"],
            "orderListId": -1,
            "clientOrderId": order_data["clientOid"],
            "price": order_data["price"],
            "origQty": order_data["size"],
            "executedQty": order_data["dealSize"],
            "cummulativeQuoteQty": order_data["dealSize"],
            "status": order_status,
            "timeInForce": order_data["timeInForce"],
            "type": order_data["type"],
            "side": order_data["side"],
            "stopPrice": order_data["stopPrice"],
            "icebergQty": order_data["visibleSize"],
            "time": order_data["createdAt"],
            "updateTime": order_data["createdAt"],
            "isWorking": order_data["isActive"],
            "origQuoteOrderQty": order_data["dealFunds"],
            "exchangeSpecific": order_data,
        }


def unify_get_symbol_orders(orders_list, futures=False, lot_size=None):
    return [unify_get_order(orderData, futures, lot_size) for orderData in orders_list]


def unify_get_best_bid_asks(ticker, symbol):
    return {
        "symbol": symbol,
        "bidPrice": ticker["bestBid"],
        "bidQty": ticker["bestBidSize"],
        "askPrice": ticker["bestAsk"],
        "askQty": ticker["bestAskSize"],
    }


def unify_get_balance_futures_out(data, is_single=False):
    if is_single:
        return {
            "asset": data["currency"],
            "free": data["availableBalance"],
            "locked": data["positionMargin"],
            "exchangeSpecific": data,
        }
    else:
        return [
            {
                "asset": assetData["currency"],
                "free": assetData["availableBalance"],
                "locked": assetData["positionMargin"],
                "exchangeSpecific": assetData,
            }
            for assetData in data
        ]


def get_futures_order_as_dict(order_data):
    params = {
        "side": order_data.side,
        "symbol": order_data.symbol,
        "type": order_data.order_type,
    }

    if "leverage" in order_data.extra_params.keys():
        params["leverage"] = order_data.extra_params["leverage"]

    if order_data.new_client_order_id is not None:
        params["clientOid"] = order_data.new_client_order_id
    else:
        params["clientOid"] = ""

    if order_data.price is not None:
        params["price"] = str(order_data.price)

    if order_data.quantity is not None:
        params["size"] = float(order_data.quantity)

    if order_data.time_in_force is not None:
        params["timeInForce"] = order_data.time_in_force

    if order_data.stop_price is not None:
        params["stopPrice"] = order_data.stop_price
        params["stop"] = order_data.extra_params["stop"]
        params["stopPriceType"] = order_data.extra_params["stopPriceType"]

    if order_data.close_position is not None:
        params["closeOrder"] = order_data.close_position

    if order_data.extra_params is not None:
        set_futures_extra_params(order_data, params)

    return params


def set_futures_extra_params(order_data, params):
    if "postOnly" in order_data.extra_params.keys():
        params["postOnly"] = order_data.extra_params["postOnly"]
    if "hidden" in order_data.extra_params.keys():
        params["hidden"] = order_data.extra_params["hidden"]
    if "iceberg" in order_data.extra_params.keys():
        params["iceberg"] = order_data.extra_params["iceberg"]
    if "visibleSize" in order_data.extra_params.keys():
        params["visibleSize"] = order_data.extra_params["visibleSize"]
    if "reduceOnly" in order_data.extra_params.keys():
        params["reduceOnly"] = order_data.extra_params["reduceOnly"]
    if "forceHold" in order_data.extra_params.keys():
        params["forceHold"] = order_data.extra_params["forceHold"]
    if "postOnly" in order_data.extra_params.keys():
        params["postOnly"] = order_data.extra_params["postOnly"]
    if "hidden" in order_data.extra_params.keys():
        params["hidden"] = order_data.extra_params["hidden"]


def unify_get_position_info(position_info):
    return {
        "entryPrice": position_info["avgEntryPrice"],
        "isAutoAddMargin": position_info["autoDeposit"],
        "leverage": position_info["realLeverage"],
        "maxNotionalValue": None,
        "liquidationPrice": position_info["liquidationPrice"],
        "markPrice": position_info["markPrice"],
        "positionAmt": position_info["currentQty"],
        "symbol": position_info["symbol"],
        "unrealizedProfit": position_info["unrealisedPnl"],
        "marginType": "cross" if position_info["crossMode"] else "isolated",
        "isolatedMargin": position_info["maintMargin"],
        "positionSide": "BOTH",
        "exchangeSpecific": position_info,
    }


def unify_get_position_infos(position_infos):
    return [unify_get_position_info(posInfo) for posInfo in position_infos]


def unify_min_trade(info, futures=False):
    if futures:
        params = {
            "minQuantity": float(info["multiplier"] * info["lotSize"]),
            "precisionStep": float(info["multiplier"] * info["lotSize"]),
            "stepPrice": info["tickSize"],
        }
        params["minQuoteQuantity"] = params["precisionStep"] * info["lastTradePrice"]
    else:
        params = {
            "minQuantity": info["baseMinSize"],
            "minQuoteQuantity": info["quoteMinSize"],
            "precisionStep": info["baseIncrement"],
            "stepPrice": info["priceIncrement"],
        }
    return params


def unify_get_income(income_list):
    return [
        {
            "symbol": income["remark"],
            "incomeType": income["type"],
            "income": income["amount"],
            "asset": income["currency"],
            "time": income["type"],
            "exchangeSpecific": income,
        }
        for income in income_list
    ]
